# -*- coding: utf-8 -*-
"""Python implementation of PVL (Parameter Value Language)."""

# Copyright 2015, 2017, 2019-2020, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import inspect
import io
import urllib.request
from pathlib import Path

from .encoder import PDSLabelEncoder, PVLEncoder
from .parser import PVLParser, OmniParser
from .collections import (
    PVLModule,
    PVLGroup,
    PVLObject,
    Quantity,
    Units,
)

__author__ = "The pvl Developers"
__email__ = "rbeyer@rossbeyer.net"
__version__ = "1.0.0"
__all__ = [
    "load",
    "loads",
    "dump",
    "dumps",
    "PVLModule",
    "PVLGroup",
    "PVLObject",
    "Quantity",
    "Units",
]


def load(path, parser=None, grammar=None, decoder=None, **kwargs):
    """Returns a Python object from parsing the file at *path*.

    :param path: an :class:`os.PathLike` which presumably has a
        PVL Module in it to parse.
    :param parser: defaults to :class:`pvl.parser.OmniParser()`.
    :param grammar: defaults to :class:`pvl.grammar.OmniGrammar()`.
    :param decoder: defaults to :class:`pvl.decoder.OmniDecoder()`.
    :param ``**kwargs``: the keyword arguments that will be passed
        to :func:`loads()` and are described there.

    If *path* is not an :class:`os.PathLike`, it will be assumed to be an
    already-opened file object, and ``.read()`` will be applied
    to extract the text.

    If the :class:`os.PathLike` or file object contains some bytes
    decodable as text, followed by some that is not (e.g. an ISIS
    cube file), that's fine, this function will just extract the
    decodable text.
    """
    return loads(
        get_text_from(path),
        parser=parser,
        grammar=grammar,
        decoder=decoder,
        **kwargs
    )


def get_text_from(path) -> str:
    try:
        p = Path(path)
        return p.read_text()
    except UnicodeDecodeError:
        # This may be the result of an ISIS cube file (or anything else)
        # where the first set of bytes might be decodable, but once the
        # image data starts, they won't be, and the above tidy function
        # fails.  So open the file as a bytestream, and read until
        # we can't decode.  We don't want to just run the .read_bytes()
        # method of Path, because this could be a giant file.
        with open(path, mode="rb") as f:
            return decode_by_char(f)
    except TypeError:
        # Not an os.PathLike, maybe it is an already-opened file object
        if path.readable():
            position = path.tell()
            try:
                s = path.read()
                if isinstance(s, bytes):
                    # Oh, it was opened in 'b' mode, need to rewind and
                    # decode.  Since the 'catch' below already does that,
                    # we'll just emit a ... contrived ... UnicodeDecodeError
                    # so we don't have to double-write the code:
                    raise UnicodeDecodeError(
                        "utf_8",
                        "dummy".encode(),
                        0,
                        1,
                        "file object in byte mode",
                    )
            except UnicodeDecodeError:
                # All of the bytes weren't decodeable, maybe the initial
                # sequence is (as above)?
                path.seek(position)  # Reset after the previous .read():
                s = decode_by_char(path)

        else:
            # Not a path, not an already-opened file.
            raise TypeError(
                "Expected an os.PathLike or an already-opened "
                "file object, but did not get either."
            )
        return s


def decode_by_char(f: io.RawIOBase) -> str:
    """Returns a ``str`` decoded from the characters in *f*.

    :param f: is expected to be a file object which has been
        opened in binary mode ('rb') or just read mode ('r').

    The *f* stream will have one character or byte at a time read from it,
    and will attempt to decode each to a string and accumulate
    those individual strings together.  Once the end of the file is found
    or an element can no longer be decoded, the accumulated string will
    be returned.
    """
    s = ""
    try:
        for elem in iter(lambda: f.read(1), b""):
            if isinstance(elem, str):
                s += elem
            else:
                s += elem.decode()
    except UnicodeError:
        # Expecting this to mean that we got to the end of decodable
        # bytes, so we're all done, and pass through to return s.
        pass

    return s


def loadu(url, parser=None, grammar=None, decoder=None, **kwargs):
    """Returns a Python object from parsing *url*.

        :param url: this will be passed to :func:`urllib.request.urlopen`
            and can be a string or a :class:`urllib.request.Request` object.
        :param parser: defaults to :class:`pvl.parser.OmniParser()`.
        :param grammar: defaults to :class:`pvl.grammar.OmniGrammar()`.
        :param decoder: defaults to :class:`pvl.decoder.OmniDecoder()`.
        :param ``**kwargs``: the keyword arguments that will be passed
            to :func:`urllib.request.urlopen` and to :func:`loads()`.

        The ``**kwargs`` will first be scanned for arguments that
        can be given to :func:`urllib.request.urlopen`.  If any are
        found, they are extracted and used.  All remaining elements
        will be passed on as keyword arguments to :func:`loads()`.

        Note that *url* can be any URL that :func:`urllib.request.urlopen`
        takes.  Certainly http and https URLs, but also file, ftp, rsync,
        sftp and more!
        """

    # Peel off the args for urlopen:
    url_args = dict()
    for a in inspect.signature(urllib.request.urlopen).parameters.keys():
        if a in kwargs:
            url_args[a] = kwargs.pop(a)

    # The object returned from urlopen will always have a .read()
    # function that returns bytes, so:
    with urllib.request.urlopen(url, **url_args) as resp:
        s = decode_by_char(resp)

    return loads(s, parser=parser, grammar=grammar, decoder=decoder, **kwargs)


def loads(s: str, parser=None, grammar=None, decoder=None, **kwargs):
    """Deserialize the string, *s*, as a Python object.

    :param s: contains some PVL to parse.
    :param parser: defaults to :class:`pvl.parser.OmniParser()`.
    :param grammar: defaults to :class:`pvl.grammar.OmniGrammar()`.
    :param decoder: defaults to :class:`pvl.decoder.OmniDecoder()`.
    :param ``**kwargs``: the keyword arguments to pass to the *parser* class
        if *parser* is none.
    """
    # decoder = __create_decoder(cls, strict, grammar=grammar, **kwargs)
    # return decoder.decode(s)

    if isinstance(s, bytes):
        # Someone passed us an old-style bytes sequence.  Although it isn't
        # a string, we can deal with it:
        s = s.decode()

    if parser is None:
        parser = OmniParser(grammar=grammar, decoder=decoder, **kwargs)
    elif not isinstance(parser, PVLParser):
        raise TypeError("The parser must be an instance of pvl.PVLParser.")

    return parser.parse(s)


def dump(module, path, **kwargs):
    """Serialize *module* as PVL text to the provided *path*.

    :param module: a ``PVLModule`` or ``dict``-like object to serialize.
    :param path: an :class:`os.PathLike`
    :param ``**kwargs``: the keyword arguments to pass to :func:`dumps()`.

    If *path* is an :class:`os.PathLike`, it will attempt to be opened
    and the serialized module will be written into that file via
    the :func:`pathlib.Path.write_text()` function, and will return
    what that function returns.

    If *path* is not an :class:`os.PathLike`, it will be assumed to be an
    already-opened file object, and ``.write()`` will be applied
    on that object to write the serialized module, and will return
    what that function returns.
    """
    try:
        p = Path(path)
        return p.write_text(dumps(module, **kwargs))

    except TypeError:
        # Not an os.PathLike, maybe it is an already-opened file object
        try:
            if isinstance(path, io.TextIOBase):
                return path.write(dumps(module, **kwargs))
            else:
                return path.write(dumps(module, **kwargs).encode())
        except AttributeError:
            # Not a path, not an already-opened file.
            raise TypeError(
                "Expected an os.PathLike or an already-opened "
                "file object for writing, but got neither."
            )


def dumps(module, encoder=None, grammar=None, decoder=None, **kwargs) -> str:
    """Returns a string where the *module* object has been serialized
    to PVL syntax.

    :param module: a ``PVLModule`` or ``dict`` like object to serialize.
    :param encoder: defaults to :class:`pvl.parser.PDSLabelEncoder()`.
    :param grammar: defaults to :class:`pvl.grammar.ODLGrammar()`.
    :param decoder: defaults to :class:`pvl.decoder.ODLDecoder()`.
    :param ``**kwargs``: the keyword arguments to pass to the encoder
        class if *encoder* is none.
    """
    if encoder is None:
        encoder = PDSLabelEncoder(grammar=grammar, decoder=decoder, **kwargs)
    elif not isinstance(encoder, PVLEncoder):
        raise TypeError("The encoder must be an instance of pvl.PVLEncoder.")

    return encoder.encode(module)



# -*- coding: utf-8 -*-
"""A program for converting PVL text to a specific PVL dialect.

The ``pvl_translate`` program will read a file with PVL text (any
of the kinds of files that :func:`pvl.load` reads) or STDIN and
will convert that PVL text to a particular PVL dialect.  It is not
particularly robust, and if it cannot make simple conversions, it
will raise errors.
"""

# Copyright 2020, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import json
import os
import sys

import pvl
from .encoder import PVLEncoder, ODLEncoder, ISISEncoder, PDSLabelEncoder


class Writer(object):
    """Base class for writers.  Descendents must implement dump().
    """

    def dump(self, dictlike: dict, outfile: os.PathLike):
        raise Exception


class PVLWriter(Writer):
    def __init__(self, encoder):
        self.encoder = encoder

    def dump(self, dictlike: dict, outfile: os.PathLike):
        return pvl.dump(dictlike, outfile, encoder=self.encoder)


class JSONWriter(Writer):
    def dump(self, dictlike: dict, outfile: os.PathLike):
        return json.dump(dictlike, outfile)


formats = dict(
    PDS3=PVLWriter(PDSLabelEncoder()),
    ODL=PVLWriter(ODLEncoder()),
    ISIS=PVLWriter(ISISEncoder()),
    PVL=PVLWriter(PVLEncoder()),
    JSON=JSONWriter(),
)


def arg_parser(formats):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-of",
        "--output_format",
        required=True,
        choices=formats.keys(),
        help="Select the format to create the new file as.",
    )
    parser.add_argument(
        "infile",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="file containing PVL text to translate, " "defaults to STDIN.",
    )
    parser.add_argument(
        "outfile",
        nargs="?",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="file to write translated PVL to, defaults " "to STDOUT.",
    )
    parser.add_argument("--version", action="version", version=pvl.__version__)
    return parser


def main():
    args = arg_parser(formats).parse_args()

    some_pvl = pvl.load(args.infile)

    formats[args.output_format].dump(some_pvl, args.outfile)
    return


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
import os
import io
import datetime
import glob
import pytest
import tempfile
import shutil

from collections import abc
from pvl.exceptions import LexerError

import pvl
from pvl import (
    PVLModule as Label,
    PVLObject as LabelObject,
    PVLGroup as LabelGroup,
)
from pvl import Quantity, Units


DATA_DIR = os.path.join(os.path.dirname(__file__), "data/")
PDS_DATA_DIR = os.path.join(os.path.dirname(__file__), "data", "pds3")
PDS_LABELS = glob.glob(os.path.join(PDS_DATA_DIR, "*.lbl"))
BROKEN_DIR = os.path.join("tests", "data", "pds3", "broken")
BAD_PDS_LABELS = glob.glob(os.path.join(BROKEN_DIR, "*.lbl"))


def test_assignment():
    label = pvl.loads("foo=bar")
    assert isinstance(label, Label)
    assert label["foo"] == "bar"

    label = pvl.loads("Group_Foo=bar")
    assert isinstance(label, Label)
    assert label["Group_Foo"] == "bar"

    label = pvl.loads("foo=bar-")
    assert isinstance(label, Label)
    assert label["foo"] == "bar-"

    label = pvl.loads("foo=bar-\n")
    assert isinstance(label, Label)
    assert label["foo"] == "bar"

    label = pvl.loads("foo=bro-\nken")
    assert isinstance(label, Label)
    assert label["foo"] == "broken"

    label = pvl.loads("foo=bro-\n ken")
    assert isinstance(label, Label)
    assert label["foo"] == "broken"


def test_spacing():
    label = pvl.loads(
        """
        foo = bar
        nospace=good
          lots_of_spacing    =    alsogood
        same = line no = problem; like=aboss
        End
    """
    )

    assert isinstance(label, Label)
    assert label["foo"] == "bar"
    assert label["nospace"] == "good"
    assert label["lots_of_spacing"] == "alsogood"
    assert label["same"] == "line"
    assert label["no"] == "problem"
    assert label["like"] == "aboss"


def test_linewrap():
    label = pvl.loads(
        """
        foo = bar-
              baz
        End
    """
    )

    assert label["foo"] == "barbaz"


def test_special():
    label = pvl.loads(
        """
        none1 = NULL
        none2 = Null
        true1 = TRUE
        true2 = True
        true3 = true
        false1 = FALSE
        false2 = False
        false3 = false
        End
    """
    )

    assert label["none1"] is None
    assert label["none2"] is None

    assert label["true1"] is True
    assert label["true2"] is True
    assert label["true3"] is True

    assert label["false1"] is False
    assert label["false2"] is False
    assert label["false3"] is False


def test_integers():
    label = pvl.loads(
        """
        integer = 42
        positive_integer = +123
        negitive_integer = -1
        invalid_integer = 1a2
        End
    """
    )

    assert isinstance(label["integer"], int)
    assert label["integer"] == 42

    assert isinstance(label["integer"], int)
    assert label["positive_integer"] == 123

    assert isinstance(label["negitive_integer"], int)
    assert label["negitive_integer"] == -1

    assert isinstance(label["invalid_integer"], str)
    assert label["invalid_integer"] == "1a2"


def test_floats():
    label = pvl.loads(
        """
        float = 1.0
        float_no_decimal = 2.
        float_no_whole = .3
        float_leading_zero = 0.5
        positive_float = +2.0
        negative_float = -1.0
        invalid_float = 1.2.3
        End
    """
    )
    assert isinstance(label["float"], float)
    assert label["float"] == 1.0

    assert isinstance(label["float_no_decimal"], float)
    assert label["float_no_decimal"] == 2.0

    assert isinstance(label["float_no_whole"], float)
    assert label["float_no_whole"] == 0.3

    assert isinstance(label["float_leading_zero"], float)
    assert label["float_leading_zero"] == 0.5

    assert isinstance(label["positive_float"], float)
    assert label["positive_float"] == 2.0

    assert isinstance(label["negative_float"], float)
    assert label["negative_float"] == -1.0

    assert isinstance(label["invalid_float"], str)
    assert label["invalid_float"] == "1.2.3"


def test_exponents():
    label = pvl.loads(
        """
        capital = -1.E-3
        lower = -1.e-3
        small = -0.45e6
        int = 31459e1
        invalid = 1e
        End
    """
    )

    assert isinstance(label["capital"], float)
    assert label["capital"] == -1.0e-3

    assert isinstance(label["lower"], float)
    assert label["lower"] == -1.0e-3

    assert isinstance(label["small"], float)
    assert label["small"] == -0.45e6

    assert isinstance(label["int"], float)
    assert label["int"] == 31459e1

    assert isinstance(label["invalid"], str)
    assert label["invalid"] == "1e"


def test_objects():
    label = pvl.loads(
        """
        Object = test_object
          foo = bar

          Object = embedded_object
            foo = bar
          End_Object

          Group = embedded_group
            foo = bar
          End_Group
        End_Object
        End
    """
    )
    test_object = label["test_object"]
    assert isinstance(test_object, LabelObject)
    assert test_object["foo"] == "bar"

    embedded_object = test_object["embedded_object"]
    assert isinstance(embedded_object, LabelObject)
    assert embedded_object["foo"] == "bar"

    embedded_group = test_object["embedded_group"]
    assert isinstance(embedded_group, LabelGroup)
    assert embedded_group["foo"] == "bar"

    with pytest.raises(LexerError):
        pvl.loads(
            """
            BEGIN_OBJECT = foo
            END_OBJECT = bar
        """
        )


def test_groups():
    label = pvl.loads(
        """
        Group = test_group
          foo = bar
          Object = embedded_object
            foo = bar
          End_Object

          Group = embedded_group
            foo = bar
          End_Group
        End_Group
        End
    """
    )
    test_group = label["test_group"]
    assert isinstance(test_group, LabelGroup)
    assert test_group["foo"] == "bar"

    embedded_object = test_group["embedded_object"]
    assert isinstance(embedded_object, LabelObject)
    assert embedded_object["foo"] == "bar"

    embedded_group = test_group["embedded_group"]
    assert isinstance(embedded_group, LabelGroup)
    assert embedded_group["foo"] == "bar"

    with pytest.raises(LexerError):
        pvl.loads(
            """
            BEGIN_GROUP = foo
            END_GROUP = bar
        """
        )


def test_alt_group_style():
    label = pvl.loads(
        """
        OBJECT               = TEST1
          FOO                = BAR
        END_OBJECT           = TEST1

        GROUP                = TEST2
          FOO                = BAR
        END_GROUP            = TEST2

        END
    """
    )
    test_group = label["TEST1"]
    assert isinstance(test_group, LabelObject)
    assert test_group["FOO"] == "BAR"

    embedded_object = label["TEST2"]
    assert isinstance(embedded_object, LabelGroup)
    assert embedded_object["FOO"] == "BAR"


def test_binary():
    label = pvl.loads(
        """
        binary_number = 2#0101#
        positive_binary_number = +2#0101#
        negative_binary_number = -2#0101#
        End
    """
    )

    assert isinstance(label["binary_number"], int)
    assert label["binary_number"] == 5

    assert isinstance(label["positive_binary_number"], int)
    assert label["positive_binary_number"] == 5

    assert isinstance(label["negative_binary_number"], int)
    assert label["negative_binary_number"] == -5

    with pytest.raises(LexerError):
        pvl.loads("empty = 2##")

    with pytest.raises(LexerError):
        pvl.loads("binary_number = 2#0101")

    with pytest.raises(LexerError):
        pvl.loads("binary_number = 2#01014201#")


def test_octal():
    label = pvl.loads(
        """
        octal_number = 8#0107#
        positive_octal_number = +8#0107#
        negative_octal_number = -8#0107#
        End
    """
    )

    assert isinstance(label["octal_number"], int)
    assert label["octal_number"] == 71

    assert isinstance(label["positive_octal_number"], int)
    assert label["positive_octal_number"] == 71

    assert isinstance(label["negative_octal_number"], int)
    assert label["negative_octal_number"] == -71

    with pytest.raises(LexerError):
        pvl.loads("empty = 8##")

    with pytest.raises(LexerError):
        pvl.loads("octal_number = 8#0107")

    with pytest.raises(LexerError):
        pvl.loads("octal_number = 8#01079#")


def test_hex():
    label = pvl.loads(
        """
        hex_number_upper = 16#100A#
        hex_number_lower = 16#100b#
        positive_hex_number = +16#100A#
        negative_hex_number = -16#100A#
        End
    """
    )

    assert isinstance(label["hex_number_upper"], int)
    assert label["hex_number_upper"] == 4106

    assert isinstance(label["hex_number_lower"], int)
    assert label["hex_number_lower"] == 4107

    assert isinstance(label["positive_hex_number"], int)
    assert label["positive_hex_number"] == 4106

    assert isinstance(label["negative_hex_number"], int)
    assert label["negative_hex_number"] == -4106

    with pytest.raises(LexerError):
        pvl.loads("empty = 16##")

    with pytest.raises(LexerError):
        pvl.loads("hex_number_upper = 16#100A")

    with pytest.raises(LexerError):
        pvl.loads("hex_number_upper = 16#100AZ#")


# I think the original 'mixed' and 'formating' tests here need
# fixing:
#
#        mixed = 'mixed"\\'quotes'
#
#        formating = "\\n\\t\\f\\v\\\\\\n\\t\\f\\v\\\\"
#
# I think they might have worked when the library was working
# with bytestrings, but these need to be modified for string
# parsing.  To be clear, the difference is in how you write things
# within a Python string and how the Python interpreter deals with
# what you write in a file like *this*.  There should be no difference
# in how the library deals with these characters in external files.
#
# For the mixed test, the single quote before 'quotes' should terminate
# the quoted string.  I think the idea here was suppsed to be that
# the \\ escaped a slash, and then that left you with the following
# two character sequence: "\'", which was meant to be interpreted as
# an 'escaped' quote which shouldn't terminate the quoted string, but
# PVL doesn't 'escape' characters, even though Python does.  So in
# the string given to the parser, Python escapes the \\ to a single
# slash (which is in the PVL unrestricted character set), followed
# by a single quote, which properly terminates the quoted string, and
# then there's an unparseable "quotes'" after it that should cause
# an error.  The test has been modified.
#
# For the formating test, it still works, but it isn't particularly
# insightful, because all of those double backslashes just escape a
# slash, so the Python string would have backslash, n, backslash, t,
# backslash, f, backslash v, backslash, backslash, backslash, etc.
# Basically just a bunch of backslashes and some letters, and no
# format effectors or space characters in it.  The formating2 test
# exercises this a little more.


def test_quotes():
    some_pvl = """
        foo = 'bar'
        empty = ''
        space = '  test  '
        double = "double'quotes"
        single = 'single"quotes'
        mixed = 'mixed"\\quotes'
        number = '123'
        date = '1918-05-11'
        multiline = 'this is a
                     multi-line string'
        continuation = "The planet Jupi-
                        ter is very big"
        formating = "\\n\\t\\f\\v\\\\\\n\\t\\f\\v\\\\"
        formating2 = "\n\t\f\v\\\n\t\f\v\\"
        End
        """

    label = pvl.loads(some_pvl)
    pvl_label = pvl.loads(some_pvl, decoder=pvl.decoder.PVLDecoder())

    assert isinstance(label["foo"], str)
    assert label["foo"] == "bar"

    assert isinstance(label["empty"], str)
    assert label["empty"] == ""

    assert isinstance(label["space"], str)
    # ODLDecoder and OmniDecoder:
    assert label["space"] == "test"
    # PVLDecoder:
    assert pvl_label["space"] == "  test  "

    assert isinstance(label["double"], str)
    assert label["double"] == "double'quotes"

    assert isinstance(label["single"], str)
    assert label["single"] == 'single"quotes'

    assert isinstance(label["mixed"], str)
    assert label["mixed"] == 'mixed"\\quotes'

    assert isinstance(label["number"], str)
    assert label["number"] == "123"

    assert isinstance(label["date"], str)
    assert label["date"] == "1918-05-11"

    # This test is really for ODL-only, PVL will retain the newline and space.
    assert isinstance(label["multiline"], str)
    assert label["multiline"] == "this is a multi-line string"

    # Also for ODL-only, but tests the OmniDecoder
    assert isinstance(label["continuation"], str)
    assert label["continuation"] == "The planet Jupiter is very big"

    assert isinstance(label["formating"], str)
    assert label["formating"] == "\\n\\t\\f\\v\\\\\\n\\t\\f\\v\\\\"
    assert label["formating2"] == "\\ \\"
    assert pvl_label["formating2"] == "\n\t\f\v\\\n\t\f\v\\"

    # The quote after the equals here starts a quoted string, but
    # it doesn't complete, and therefore there's no assignment to foo.
    with pytest.raises(LexerError):
        pvl.loads('foo = "bar')

    with pytest.raises(LexerError):
        pvl.loads("foo = 'bar")

    # The \b character is not an allowed PVL character.
    with pytest.raises(LexerError):
        # pvl.loads("foo = '\\bar'")
        pvl.loads("foo = '\bar'")


def test_comments():
    some_pvl = """
        /* comment on line */
        # here is a line comment
        /* here is a multi-
        line comment */
        foo = bar /* comment at end of line */
        weird/* in the */=/*middle*/comments
        baz = bang # end line comment
        End
    """

    # Strict PVL doesn't allow #-comments
    with pytest.raises(LexerError):
        pvl.loads(some_pvl, grammar=pvl.grammar.PVLGrammar())

    # But the OmniGrammar does:
    label = pvl.loads(some_pvl)

    assert len(label) == 3

    assert isinstance(label["foo"], str)
    assert label["foo"] == "bar"

    assert isinstance(label["foo"], str)
    assert label["weird"] == "comments"

    assert isinstance(label["foo"], str)
    assert label["baz"] == "bang"

    with pytest.raises(LexerError):
        pvl.loads(b"/*")


def test_dates():
    some_pvl = """
        date1          = 1990-07-04
        date2          = 1990-158
        date3          = 2001-001
        date4          = 2001-01-01
        time1          = 12:00
        time_s         = 12:00:45
        time_s_float   = 12:00:45.4571
        time_tz1       = 15:24:12Z
        time_tz2       = 01:12:22+07
        time_tz3       = 01:12:22+7
        time_tz4       = 01:10:39.4575+07
        datetime1      = 1990-07-04T12:00
        datetime2      = 1990-158T15:24:12Z
        datetime3      = 2001-001T01:10:39+7
        datetime4      = 2001-001T01:10:39.457591+7
        End
    """

    # PVL doesn't allow numeric TZ offsets, like "+07"
    with pytest.raises(LexerError):
        pvl.loads(some_pvl, decoder=pvl.decoder.PVLDecoder())

    # But ODL, and the OmniDecoder do:
    label = pvl.loads(some_pvl)

    tz_plus_7 = datetime.timezone(datetime.timedelta(hours=7))
    utc = datetime.timezone.utc

    assert isinstance(label["date1"], datetime.date)
    assert label["date1"] == datetime.date(1990, 7, 4)

    assert isinstance(label["date2"], datetime.date)
    assert label["date2"] == datetime.date(1990, 6, 7)

    assert isinstance(label["date3"], datetime.date)
    assert label["date3"] == datetime.date(2001, 1, 1)

    assert isinstance(label["date4"], datetime.date)
    assert label["date4"] == datetime.date(2001, 1, 1)

    assert isinstance(label["time1"], datetime.time)
    assert label["time1"] == datetime.time(12, tzinfo=utc)

    assert isinstance(label["time_s"], datetime.time)
    assert label["time_s"] == datetime.time(12, 0, 45, tzinfo=utc)

    assert isinstance(label["time_s_float"], datetime.time)
    assert label["time_s_float"] == datetime.time(
        12, 0, 45, 457100, tzinfo=utc
    )

    assert isinstance(label["time_tz1"], datetime.time)
    assert label["time_tz1"] == datetime.time(15, 24, 12, tzinfo=utc)

    assert isinstance(label["time_tz2"], datetime.time)
    assert label["time_tz2"] == datetime.time(1, 12, 22, tzinfo=tz_plus_7)

    assert isinstance(label["datetime1"], datetime.datetime)
    assert label["datetime1"] == datetime.datetime(1990, 7, 4, 12, tzinfo=utc)

    assert isinstance(label["datetime2"], datetime.datetime)
    assert label["datetime2"] == datetime.datetime(
        1990, 6, 7, 15, 24, 12, tzinfo=utc
    )

    assert isinstance(label["time_tz3"], datetime.time)
    assert label["time_tz3"] == datetime.time(1, 12, 22, tzinfo=tz_plus_7)

    assert isinstance(label["time_tz4"], datetime.time)
    assert label["time_tz4"] == datetime.time(
        1, 10, 39, 457500, tzinfo=tz_plus_7
    )

    assert isinstance(label["datetime3"], datetime.datetime)
    assert label["datetime3"] == datetime.datetime(
        2001, 1, 1, 1, 10, 39, tzinfo=tz_plus_7
    )

    assert isinstance(label["datetime4"], datetime.datetime)
    assert label["datetime4"] == datetime.datetime(
        2001, 1, 1, 1, 10, 39, 457591, tzinfo=tz_plus_7
    )


def test_set():
    label = pvl.loads(
        """
        strings = {a, b, c}
        nospace={a,b,c}
        numbers = {1, 2, 3}
        mixed = {a, 1, 2.5}
        multiline = {a,
                     b,
                     c}
        empty = {}
        End
    """
    )

    # sets are non-hashable, so pvl is now returning frozensets
    # assert isinstance(label['strings'], set)
    assert isinstance(label["strings"], frozenset)
    assert len(label["strings"]) == 3
    assert "a" in label["strings"]
    assert "b" in label["strings"]
    assert "c" in label["strings"]

    # assert isinstance(label['nospace'], set)
    assert isinstance(label["nospace"], frozenset)
    assert len(label["nospace"]) == 3
    assert "a" in label["nospace"]
    assert "b" in label["nospace"]
    assert "c" in label["nospace"]

    # assert isinstance(label['numbers'], set)
    assert isinstance(label["numbers"], frozenset)
    assert len(label["numbers"]) == 3
    assert 1 in label["numbers"]
    assert 2 in label["numbers"]
    assert 3 in label["numbers"]

    # assert isinstance(label['mixed'], set)
    assert isinstance(label["mixed"], frozenset)
    assert len(label["mixed"]) == 3
    assert "a" in label["mixed"]
    assert 1 in label["mixed"]
    assert 2.5 in label["mixed"]

    # assert isinstance(label['multiline'], set)
    assert isinstance(label["multiline"], frozenset)
    assert len(label["multiline"]) == 3
    assert "a" in label["multiline"]
    assert "b" in label["multiline"]
    assert "c" in label["multiline"]

    # assert isinstance(label['empty'], set)
    assert isinstance(label["empty"], frozenset)
    assert len(label["empty"]) == 0


def test_sequence():
    label = pvl.loads(
        """
        strings = (a, b, c)
        nospace=(a,b,c)
        numbers = (1, 2, 3)
        mixed = (a, 1, 2.5)
        empty = ()
        multiline = (a,
                     b,
                     c)
        linewrap = (1.234,1.2-
                    34,1.234-
                    ,1.234)
        End
    """
    )

    assert isinstance(label["strings"], list)
    assert len(label["strings"]) == 3
    assert label["strings"][0] == "a"
    assert label["strings"][1] == "b"
    assert label["strings"][2] == "c"

    assert isinstance(label["nospace"], list)
    assert len(label["nospace"]) == 3
    assert label["nospace"][0] == "a"
    assert label["nospace"][1] == "b"
    assert label["nospace"][2] == "c"

    assert isinstance(label["numbers"], list)
    assert len(label["numbers"]) == 3
    assert label["numbers"][0] == 1
    assert label["numbers"][1] == 2
    assert label["numbers"][2] == 3

    assert isinstance(label["mixed"], list)
    assert len(label["mixed"]) == 3
    assert label["mixed"][0] == "a"
    assert label["mixed"][1] == 1
    assert label["mixed"][2] == 2.5

    assert isinstance(label["empty"], list)
    assert len(label["empty"]) == 0

    assert isinstance(label["multiline"], list)
    assert len(label["multiline"]) == 3
    assert label["multiline"][0] == "a"
    assert label["multiline"][1] == "b"
    assert label["multiline"][2] == "c"

    assert isinstance(label["linewrap"], list)
    assert len(label["linewrap"]) == 4
    assert label["linewrap"][0] == 1.234
    assert label["linewrap"][1] == 1.234
    assert label["linewrap"][2] == 1.234
    assert label["linewrap"][3] == 1.234


def test_sequence_backslashes():
    # We need to escape the slashes here within Python:
    some_pvl = """SPICE_FILE_NAME = ("sclk\\ROS_160929_STEP.TSC",
    "lsk\\NAIF0011.TLS", "fk\\ROS_V26.TF", "ik\\ROS_OSIRIS_V13.TI",
    "spk\\RORB_DV_145_01_______00216.BSP", "ck\\RATT_DV_145_01_01____00216.BC",
    "pck\\PCK00010.TPC", "spk\\DE405.BSP", "pck\\ROS_CGS_RSOC_V03.TPC",
    "fk\\ROS_CHURYUMOV_V01.TF", "spk\\CORB_DV_145_01_______00216.BSP",
    "ck\\CATT_DV_145_01_______00216.BC")"""

    sclk_str = "sclk\\ROS_160929_STEP.TSC"

    p1 = pvl.loads(some_pvl)
    p1["SPICE_FILE_NAME"][0] = sclk_str

    # But they read in just fine without
    backslashes_lbl = os.path.join(PDS_DATA_DIR, "backslashes.lbl")
    p2 = pvl.load(backslashes_lbl)
    p2["SPICE_FILE_NAME"][0] = sclk_str


def test_units():
    label = pvl.loads(
        """
        foo = 42 <beards>
        g = 9.8 <m/s>
        list = (1, 2, 3) <numbers>
        cool = (1 <number>)
        End
    """
    )
    assert isinstance(label["foo"], Quantity)
    assert label["foo"].value == 42
    assert label["foo"].units == "beards"

    assert isinstance(label["g"], Quantity)
    assert label["g"].value == 9.8
    assert label["g"].units == "m/s"

    assert isinstance(label["list"], Quantity)
    assert isinstance(label["list"].value, list)
    assert label["list"].units == "numbers"

    assert isinstance(label["cool"], list)
    assert isinstance(label["cool"][0], Quantity)
    assert label["cool"][0].value == 1
    assert label["cool"][0].units == "number"

    with pytest.raises(LexerError):
        pvl.loads(b"foo = bar <")


def test_delimiters():
    label = pvl.loads(
        """
        foo = 1;
        Object = embedded_object;
          foo = bar;
        End_Object;
        bar = 2;
        Group = embedded_group;
          foo = bar;
        End_Group;
        End;
    """
    )

    assert isinstance(label, Label)
    assert label["foo"] == 1
    assert label["bar"] == 2

    assert isinstance(label["embedded_object"], LabelObject)
    assert label["embedded_object"]["foo"] == "bar"

    assert isinstance(label["embedded_group"], LabelGroup)
    assert label["embedded_group"]["foo"] == "bar"


def test_isis_output():
    label = pvl.load(os.path.join(DATA_DIR, "isis_output.txt"))
    assert label["Results"]["TotalPixels"] == 2048000

    naif = pvl.load(os.path.join(DATA_DIR, "isis_naif.txt"))
    assert naif["NaifKeywords"]["INS-143400_LIGHTTIME_CORRECTION"] == "LT+S"


def test_cube_label():
    with open(os.path.join(DATA_DIR, "pattern.cub"), "rb") as fp:
        label = pvl.load(fp)

    assert isinstance(label["Label"], abc.Mapping)
    assert label["Label"]["Bytes"] == 65536

    assert isinstance(label["IsisCube"], abc.Mapping)
    assert isinstance(label["IsisCube"]["Core"], abc.Mapping)
    assert label["IsisCube"]["Core"]["StartByte"] == 65537
    assert label["IsisCube"]["Core"]["Format"] == "Tile"
    assert label["IsisCube"]["Core"]["TileSamples"] == 128
    assert label["IsisCube"]["Core"]["TileLines"] == 128

    assert isinstance(label["IsisCube"]["Core"]["Dimensions"], abc.Mapping)
    assert label["IsisCube"]["Core"]["Dimensions"]["Samples"] == 90
    assert label["IsisCube"]["Core"]["Dimensions"]["Lines"] == 90
    assert label["IsisCube"]["Core"]["Dimensions"]["Bands"] == 1

    assert isinstance(label["IsisCube"]["Core"]["Pixels"], abc.Mapping)
    assert label["IsisCube"]["Core"]["Pixels"]["Type"] == "Real"
    assert label["IsisCube"]["Core"]["Pixels"]["ByteOrder"] == "Lsb"
    assert label["IsisCube"]["Core"]["Pixels"]["Base"] == 0.0
    assert label["IsisCube"]["Core"]["Pixels"]["Multiplier"] == 1.0


def test_cube_label_r():
    with open(os.path.join(DATA_DIR, "pattern.cub"), "r") as fp:
        label = pvl.load(fp)

    assert isinstance(label["Label"], abc.Mapping)
    assert label["Label"]["Bytes"] == 65536


def test_pds3_sample_image():
    infile = os.path.join(PDS_DATA_DIR, "simple_image_1.lbl")
    label = pvl.load(infile)
    assert label["RECORD_TYPE"] == "FIXED_LENGTH"
    assert label["RECORD_BYTES"] == 824
    assert label["LABEL_RECORDS"] == 1
    assert label["FILE_RECORDS"] == 601
    assert label["IMAGE"]["LINES"] == 600
    assert label["IMAGE"]["LINE_SAMPLES"] == 824
    image_group = label["IMAGE"]
    assert image_group["SAMPLE_TYPE"] == "MSB_INTEGER"
    assert image_group["SAMPLE_BITS"] == 8
    assert abs(image_group["MEAN"] - 51.6778539644) <= 0.00001
    assert image_group["MEDIAN"] == 50.0
    assert image_group["MINIMUM"] == 0
    assert image_group["MAXIMUM"] == 255
    assert image_group["STANDARD_DEVIATION"] == 16.97019
    assert image_group["CHECKSUM"] == 25549531


def test_load_all_sample_labels():
    for filename in PDS_LABELS:
        label = pvl.load(filename)
        assert isinstance(label, Label)


def test_unicode():
    label = pvl.loads(u"foo=bar")
    assert isinstance(label, Label)
    assert label["foo"] == "bar"


def test_bytes():
    label = pvl.loads(b"foo=bar")
    assert isinstance(label, Label)
    assert label["foo"] == "bar"


def test_end_comment():
    label = pvl.loads(b"END/* commnet */")
    assert isinstance(label, Label)
    assert len(label) == 0


def test_parse_error():
    with pytest.raises(pvl.parser.ParseError):
        pvl.loads(b"foo=", parser=pvl.PVLParser())

    with pytest.raises(LexerError):
        pvl.loads(b"=")

    with pytest.raises(LexerError):
        pvl.loads(b"(}")

    # Identical with above?
    # with pytest.raises(pvl.decoder.ParseError):
    #     pvl.loads(b'foo=')

    with pytest.raises(LexerError):
        pvl.loads(b"foo=!")

    with pytest.raises(pvl.parser.ParseError):
        pvl.loads(b"foo")

    with pytest.raises(pvl.parser.ParseError):
        pvl.load(io.BytesIO(b"foo"))


EV = pvl.parser.EmptyValueAtLine


@pytest.mark.parametrize(
    "label, expected, expected_errors",
    [
        (
            "broken1.lbl",
            [("foo", "bar"), ("life", EV(2)), ("monty", "python")],
            [2],
        ),
        ("broken2.lbl", [("foo", "bar"), ("life", EV(2))], [2]),  # ParseError
        ("broken3.lbl", [("foo", EV(1)), ("life", 42)], [1]),
        (
            "broken4.lbl",
            [("foo", "bar"), ("life", EV(2)), ("monty", EV(3))],
            [2, 3],
        ),
        (
            "broken5.lbl",
            [("foo", EV(1)), ("life", EV(2)), ("monty", "python")],
            [1, 2],
        ),
        (
            "broken6.lbl",
            [("foo", EV(1)), ("life", EV(1)), ("monty", EV(1))],
            [1, 2, 3],
        ),
        (
            "broken7.lbl",
            [
                ("foo", 1),
                (
                    "embedded_object",
                    pvl.PVLObject([("foo", "bar"), ("life", EV(1))]),
                ),
            ],
            [4],
        ),
        (
            "broken8.lbl",
            [
                ("foo", 1),
                (
                    "embedded_group",
                    pvl.PVLGroup([("foo", "bar"), ("life", EV(1))]),
                ),
            ],
            [4],
        ),
        ("broken9.lbl", [("foo", 42), ("bar", EV(1))], [2]),
        ("broken10.lbl", [("foo", Units(42, "beards")), ("cool", EV(1))], [2]),
        (
            "broken11.lbl",
            [("foo", EV(1)), ("cool", [Units(1, "beards")])],
            [1],
        ),
        (
            "broken12.lbl",
            [
                ("strs", ["a", "b"]),
                ("empty", EV(2)),
                ("multiline", ["a", "b"]),
            ],
            [2],
        ),
        (
            "broken13.lbl",
            [
                ("same", "line"),
                ("no", "problem"),
                ("foo", EV(1)),
                ("bar", EV(2)),
            ],
            [1, 2],
        ),
        (
            "broken14.lbl",
            [("foo", "bar"), ("weird", EV(3)), ("baz", "bang")],
            [3],
        ),
        (
            "broken15.lbl",
            [("foo", "bar"), ("weird", "comment"), ("baz", EV(4))],
            [4],
        ),
        (
            "broken16.lbl",
            [("foo", EV(2)), ("weird", "comment"), ("baz", "bang")],
            [2],
        ),
    ],
)
def test_broken_labels(label, expected, expected_errors):
    with open(os.path.join(BROKEN_DIR, label), "rb") as stream:
        module = pvl.load(stream)

    expected = pvl.PVLModule(expected)

    # This isn't a deep comparison, since all EmptyValueAtLine
    # are empty strings, regardless of their .lineno value, they
    # always compare equal.
    assert module == expected

    # But this should compare the 'line numbers' of the EmptyValueAtLine
    # objects.
    assert module.errors == expected_errors


@pytest.mark.parametrize(
    "label",
    [
        "broken1.lbl",
        # 'broken2.lbl',  # moved to test_broken_labels_ParseError
        "broken3.lbl",
        "broken4.lbl",
        "broken5.lbl",
        "broken6.lbl",
        "broken7.lbl",
        "broken8.lbl",
        "broken9.lbl",
        "broken10.lbl",
        "broken11.lbl",
        "broken12.lbl",
        "broken13.lbl",
        "broken14.lbl",
        "broken15.lbl",
        "broken16.lbl",
    ],
)
def test_broken_labels_LexerError(label):
    with open(os.path.join(BROKEN_DIR, label), "rb") as stream:
        # with pytest.raises(pvl.decoder.ParseError):
        with pytest.raises(LexerError):
            pvl.load(stream, parser=pvl.PVLParser())


def test_broken_labels_ParseError():
    with open(os.path.join(BROKEN_DIR, "broken2.lbl"), "rb") as stream:
        with pytest.raises(pvl.parser.ParseError):
            pvl.load(stream, parser=pvl.PVLParser())


def test_EmptyValueAtLine():
    test_ev = pvl.parser.EmptyValueAtLine(1)
    assert test_ev == ""
    assert "foo" + test_ev == "foo"
    assert isinstance(test_ev, str)
    assert test_ev.lineno == 1
    assert int(test_ev) == 0
    assert float(test_ev) == 0.0
    trep = (
        "EmptyValueAtLine(1 does not have a value. Treat as an empty string)"
    )
    assert repr(test_ev) == trep


def test_load_all_bad_sample_labels():
    for filename in BAD_PDS_LABELS:
        label = pvl.load(filename)
        assert isinstance(label, Label)


# Below here are tests that deal with exercising both the decoding and
# encoding, and mostly rely on equivalence of the initially loaded
# PVLModule, and the PVLModule loaded from the dumped PVLModule.
#
# The pre-1.0 pvl ensured that all of the 'PVL' in the PDS_LABELS files
# was interchangeably readable and writable.  The thing is that the
# 'PVL' contained in those files doesn't all conform to PDS standards.
# So now that the library properly distinguishes between the three
# 'flavors' of PVL: actual PVL, ODL, and the subset of ODL that the
# PDS accepts, all of the files can be ingested (as demonstrated in
# the tests above), but some of that PVL cannot be written out by the
# PDSLabelEncoder (which is the most strict).  Or is altered by the
# encoder (labels are uppercased in ODL and PDS).
#
# So we must divide the PDS_LABELS into those that represent PVL
# that can be written out by the default PDSLabelEncoder, and those
# that must be written by the slightly more permissive ODLEncoder,
# and those that must be written by the PVLEncoder, so that when
# they are read back in by the OmniParser, the Python objects can
# compare true for these tests to pass.


PDS_COMPLIANT = list()
ODL_COMPLIANT = list()
PVL_COMPLIANT = list()

for filename in PDS_LABELS:
    if (
        "tests/data/pds3/set1.lbl" in filename
        or "tests/data/pds3/set2.lbl" in filename  # float in set
        or "tests/data/pds3/dates.lbl" in filename  # float in set
    ):  # tz in dates
        ODL_COMPLIANT.append(filename)
    elif "tests/data/pds3/float1.lbl" in filename:  # lowercase in label
        PVL_COMPLIANT.append(filename)  # FlOAT not FLOAT
    else:
        PDS_COMPLIANT.append(filename)


def test_dump_stream():
    for filename in PDS_COMPLIANT:
        # print(filename)
        label = pvl.load(filename)
        # print(label)
        stream = io.BytesIO()
        pvl.dump(label, stream)
        stream.seek(0)
        assert label == pvl.load(stream)

    for filename in ODL_COMPLIANT:
        # print(filename)
        label = pvl.load(filename)
        # print(label)
        stream = io.BytesIO()
        pvl.dump(label, stream, encoder=pvl.encoder.ODLEncoder())
        stream.seek(0)
        assert label == pvl.load(stream)

        with pytest.raises(ValueError):
            pvl.dump(label, stream)

    for filename in PVL_COMPLIANT:
        # print(filename)
        label = pvl.load(filename)
        # print(label)
        stream = io.BytesIO()
        pvl.dump(label, stream, encoder=pvl.encoder.PVLEncoder())
        stream.seek(0)
        assert label == pvl.load(stream)


def test_dump_to_file():
    tmpdir = tempfile.mkdtemp()

    try:
        for filename in PDS_COMPLIANT:
            label = pvl.load(filename)
            tmpfile = os.path.join(tmpdir, os.path.basename(filename))
            pvl.dump(label, tmpfile)
            assert label == pvl.load(tmpfile)
    finally:
        shutil.rmtree(tmpdir)


def test_default_encoder():
    for filename in PDS_COMPLIANT:
        label = pvl.load(filename)
        assert label == pvl.loads(pvl.dumps(label))


# the IsisCubeLabelEncoder class is deprecated
# def test_cube_encoder():
#     for filename in PDS_COMPLIANT:
#         label = pvl.load(filename)
#         encoder = pvl.encoder.IsisCubeLabelEncoder
#         assert label == pvl.loads(pvl.dumps(label, cls=encoder))


def test_pds_encoder():
    for filename in PDS_COMPLIANT:
        label = pvl.load(filename)
        encoder = pvl.encoder.PDSLabelEncoder()
        assert label == pvl.loads(pvl.dumps(label, encoder=encoder))


def test_special_values():
    module = pvl.PVLModule(
        [("bool_true", True), ("bool_false", False), ("null", None),]
    )
    assert module == pvl.loads(
        pvl.dumps(module, encoder=pvl.encoder.PVLEncoder())
    )

    # IsisCubeLabelEncoder class is deprecated
    # encoder = pvl.encoder.IsisCubeLabelEncoder
    # assert module == pvl.loads(pvl.dumps(module, cls=encoder))

    # This is now the default:
    # encoder = pvl.encoder.PDSLabelEncoder

    # But to compare with PVL written by the PDSLabelEncoder, the
    # labels must be uppercase:
    pds_module = pvl.PVLModule(
        [("BOOL_TRUE", True), ("BOOL_FALSE", False), ("NULL", None),]
    )
    assert pds_module == pvl.loads(pvl.dumps(module))


def test_special_strings():
    module = pvl.PVLModule(
        [
            ("single_quote", "'"),
            ("double_quote", '"'),
            # ('mixed_quotes', '"\''),  # see above about escaped quotes.
        ]
    )
    assert module == pvl.loads(
        pvl.dumps(module, encoder=pvl.encoder.PVLEncoder())
    )

    # IsisCubeLabelEncoder class is deprecated
    # encoder = pvl.encoder.IsisCubeLabelEncoder
    # assert module == pvl.loads(pvl.dumps(module, cls=encoder))

    # This just duplicates the above test
    # encoder = pvl.encoder.PDSLabelEncoder
    # assert module == pvl.loads(pvl.dumps(module, cls=encoder))


def test_unkown_value():
    class UnknownType(object):
        pass

    with pytest.raises(TypeError):
        print(pvl.dumps({"foo": UnknownType()}))


def test_quoated_strings():
    module = pvl.PVLModule(
        [
            ("int_like", "123"),
            ("float_like", ".2"),
            ("date", "1987-02-25"),
            ("time", "03:04:05"),
            ("datetime", "1987-02-25T03:04:05"),
            ("keyword", "END"),
            # both kinds of quotes aren't allowed:
            # ('restricted_chars', '&<>\'{},[]=!#()%";|'),
            ("restricted_chars", '&<>{},[]=!#()%";|'),
            ("restricted_seq", "/**/"),
        ]
    )
    assert module == pvl.loads(
        pvl.dumps(module, encoder=pvl.encoder.PVLEncoder())
    )

    # IsisCubeLabelEncoder class is deprecated
    # encoder = pvl.encoder.IsisCubeLabelEncoder
    # assert module == pvl.loads(pvl.dumps(module, cls=encoder))

    # This just duplicates the above test.
    # encoder = pvl.encoder.PDSLabelEncoder
    # assert module == pvl.loads(pvl.dumps(module, cls=encoder))


def test_dump_to_file_insert_before():
    tmpdir = tempfile.mkdtemp()

    try:
        for filename in PDS_COMPLIANT:
            label = pvl.load(filename)
            if os.path.basename(filename) != "empty.lbl":
                label.insert_before("PDS_VERSION_ID", [("new", "item")])
            tmpfile = os.path.join(tmpdir, os.path.basename(filename))
            pvl.dump(label, tmpfile, encoder=pvl.encoder.PVLEncoder())
            assert label == pvl.load(tmpfile)
    finally:
        shutil.rmtree(tmpdir)


def test_dump_to_file_insert_after():
    tmpdir = tempfile.mkdtemp()

    try:
        for filename in PDS_COMPLIANT:
            label = pvl.load(filename)
            if os.path.basename(filename) != "empty.lbl":
                label.insert_after("PDS_VERSION_ID", [("new", "item")])
            tmpfile = os.path.join(tmpdir, os.path.basename(filename))
            pvl.dump(label, tmpfile, encoder=pvl.encoder.PVLEncoder())
            assert label == pvl.load(tmpfile)
    finally:
        shutil.rmtree(tmpdir)



# -*- coding: utf-8 -*-

# Copyright 2019-2020, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.


from .decoder import PVLDecoder
from .grammar import PVLGrammar


class Token(str):
    """A PVL-aware string.

    :var content: A string that is the Token text.

    :var grammar: A pvl.grammar object, if None or not specified, it will
                  be set to the grammar parameter of *decoder* (if
                  *decoder* is not None) or will default to PVLGrammar().

    :var decoder: A pvl.decoder object, defaults to
                  PVLDecoder(grammar=*grammar*).

    :var pos: Integer that describes the starting position of this
              Token in the source string, defaults to zero.
    """

    def __new__(cls, content, grammar=None, decoder=None, pos=0):
        return str.__new__(cls, content)

    def __init__(self, content, grammar=None, decoder=None, pos=0):
        if grammar is None:
            if decoder is not None:
                self.grammar = decoder.grammar
            else:
                self.grammar = PVLGrammar()
        elif isinstance(grammar, PVLGrammar):
            self.grammar = grammar
        else:
            raise TypeError("The grammar object is not of type PVLGrammar.")

        if decoder is None:
            self.decoder = PVLDecoder(grammar=self.grammar)
        elif isinstance(decoder, PVLDecoder):
            self.decoder = decoder
        else:
            raise TypeError("The decoder object is not of type PVLDecoder.")

        self.pos = pos

    def __repr__(self):
        return f"{self.__class__.__name__}('{self}', " f"'{self.grammar}')"

    def __index__(self):
        return self.decoder.decode_non_decimal(str(self))

    def __float__(self):
        return self.decoder.decode_decimal(str(self))

    def split(self, sep=None, maxsplit=-1) -> list:
        """Extends ``str.split()`` that calling split() on a Token
        returns a list of Tokens.
        """
        str_list = super().split(sep, maxsplit)
        tkn_list = list()
        for t in str_list:
            tkn_list.append(
                Token(t, grammar=self.grammar, decoder=self.decoder)
            )
        return tkn_list

    def replace(self, *args):
        """Extends ``str.replace()`` to return a Token."""
        return Token(
            super().replace(*args), grammar=self.grammar, decoder=self.decoder
        )

    def lstrip(self, chars=None):
        """Extends ``str.lstrip()`` to strip whitespace according
        to the definition of whitespace in the Token's grammar
        instead of the default Python whitespace definition.
        """
        return self._strip(super().lstrip, chars)

    def rstrip(self, chars=None):
        """Extends ``str.rstrip()`` to strip whitespace according
        to the definition of whitespace in the Token's grammar
        instead of the default Python whitespace definition.
        """
        return self._strip(super().rstrip, chars)

    def strip(self, chars=None):
        """Extends ``str.strip()`` to strip whitespace according
        to the definition of whitespace in the Token's grammar
        instead of the default Python whitespace definition.
        """
        return self._strip(super().strip, chars)

    def _strip(self, strip_func, chars=None):
        # Shared functionality for the various strip functions.
        if chars is None:
            chars = "".join(self.grammar.whitespace)
        return Token(
            strip_func(chars), grammar=self.grammar, decoder=self.decoder
        )

    def isspace(self) -> bool:
        """Overrides ``str.isspace()`` to be the same as Token's
        is_space() function, so that we don't get inconsisent
        behavior if someone forgets an underbar.
        """
        # So that we don't get inconsisent behavior
        # if someone forgets an underbar.
        return self.is_space()

    def is_space(self) -> bool:
        """Return true if the Token contains whitespace according
        to the definition of whitespace in the Token's grammar
        and there is at least one character, false otherwise.
        """
        if len(self) == 0:
            return False

        return all(c in self.grammar.whitespace for c in self)

    def is_WSC(self) -> bool:
        """Return true if the Token is white space characters or comments
        according to the Token's grammar, false otherwise.
        """
        if self.is_comment():
            return True

        if self.is_space():
            return True

        for ws in reversed(self.grammar.whitespace):
            temp = self.replace(ws, " ")

        return all(t.is_comment() for t in temp.split())

    def is_comment(self) -> bool:
        """Return true if the Token is a comment according to the
        Token's grammar (defined as beginning and ending with
        comment delimieters), false otherwise.
        """
        for pair in self.grammar.comments:
            if self.startswith(pair[0]) and self.endswith(pair[1]):
                return True
        return False

    def is_quote(self) -> bool:
        """Return true if the Token is a comment character (or
        multicharacter comment delimiter) according to the
        Token's grammar, false otherwise.
        """
        if self in self.grammar.quotes:
            return True
        else:
            return False

    def is_quoted_string(self) -> bool:
        """Return true if the Token can be converted to a quoted
        string by the Token's decoder, false otherwise.
        """
        try:
            self.decoder.decode_quoted_string(self)
            return True
        except ValueError:
            return False

    def is_delimiter(self) -> bool:
        """Return true if the Token is a delimiter character
        (e.g. the ';' in PVL) according to the Token's grammar,
        false otherwise.
        """
        if self in self.grammar.delimiters:
            return True
        return False

    def is_begin_aggregation(self) -> bool:
        """Return true if the Token is a begin aggregation
        keyword (e.g. 'BEGIN_GROUP' in PVL) according to
        the Token's grammar, false otherwise.
        """
        for k in self.grammar.aggregation_keywords.keys():
            if self.casefold() == k.casefold():
                return True
        return False

    def is_unquoted_string(self) -> bool:
        """Return false if the Token has any
        reserved characters, comment characters, whitespace
        characters or could be interpreted as a number,
        date, or time according to the Token's grammar,
        true otherwise.
        """
        for char in self.grammar.reserved_characters:
            if char in self:
                return False

        for pair in self.grammar.comments:
            if pair[0] in self:
                return False
            if pair[1] in self:
                return False

        if self.is_numeric() or self.is_datetime():
            return False

        for char in self.grammar.whitespace:
            if char in self:
                return False

        return True

    def is_string(self) -> bool:
        """Return true if either the Token's is_quoted_string()
        or is_unquoted_string() return true, false otherwise.
        """
        if self.is_quoted_string() or self.is_unquoted_string():
            return True
        return False

    def is_parameter_name(self) -> bool:
        """Return true if the Token is an unquoted string that
        isn't a reserved_keyword according to the Token's
        grammar, false otherwise.
        """
        for word in self.grammar.reserved_keywords:
            if word.casefold() == self.casefold():
                return False

        return self.is_unquoted_string()

    def is_end_statement(self) -> bool:
        """Return true if the Token matches an end statement
        from its grammar, false otherwise.
        """
        for e in self.grammar.end_statements:
            if e.casefold() == self.casefold():
                return True
        return False

    def isnumeric(self) -> bool:
        """Overrides ``str.isnumeric()`` to be the same as Token's
        is_numeric() function, so that we don't get inconsisent behavior
        if someone forgets an underbar.
        """
        return self.is_numeric()

    def is_numeric(self) -> bool:
        """Return true if the Token's is_decimal() or is_non_decimal()
        functions return true, false otherwise.
        """
        if self.is_decimal() or self.is_non_decimal():
            return True

        return False

    def is_decimal(self) -> bool:
        """Return true if the Token's decoder can convert the Token
        to a decimal value, false otherwise.
        """
        try:
            self.decoder.decode_decimal(self)
            return True
        except ValueError:
            return False

    def is_non_decimal(self) -> bool:
        """Return true if the Token's decoder can convert the Token
        to a numeric non-decimal value, false otherwise.
        """
        try:
            self.decoder.decode_non_decimal(self)
            return True
        except ValueError:
            return False

    # Took these out, since some grammars allow a much wider
    # range of radix values.
    #
    # def is_binary(self) -> bool:
    #     if self.grammar.binary_re.fullmatch(self) is None:
    #         return False
    #     else:
    #         return True

    # def is_octal(self) -> bool:
    #     if self.grammar.octal_re.fullmatch(self) is None:
    #         return False
    #     else:
    #         return True

    # def is_hex(self) -> bool:
    #     if self.grammar.hex_re.fullmatch(self) is None:
    #         return False
    #     else:
    #         return True

    def is_datetime(self) -> bool:
        """Return true if the Token's decoder can convert the Token
        to a datetime, false otherwise.

        Separate is_date() or is_time() functions aren't needed,
        since PVL parsing doesn't distinguish between them.
        If a user needs that distinction the decoder's
        decode_datetime(self) function should return a datetime
        time, date, or datetime object, as appropriate, and
        a user can use isinstance() to check.
        """
        try:
            self.decoder.decode_datetime(self)
            return True
        except ValueError:
            return False

    def is_simple_value(self) -> bool:
        """Return true if the Token's decoder can convert the Token
        to a 'simple value', however the decoder defines that, false
        otherwise.
        """
        try:
            self.decoder.decode_simple_value(self)
            return True
        except ValueError:
            return False



# -*- coding: utf-8 -*-
"""This module has tests for the pvl collections module."""

# Copyright 2015, 2017, 2019-2020, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from collections import abc
import unittest

import pvl
from pvl.collections import OrderedMultiDict


class DictLike(abc.Mapping):
    def __init__(self):
        self.list = ["a", "b", "a"]

    def __getitem__(self, key):
        return 42

    def __iter__(self):
        return iter(self.list)

    def __len__(self):
        return len(self.list)


class TestMultiDicts(unittest.TestCase):
    def setUp(self):
        try:
            from pvl.collections import PVLMultiDict

            self.classes = (OrderedMultiDict, PVLMultiDict)
        except ImportError:
            self.classes = (OrderedMultiDict,)

    def test_empty(self):
        for cls in self.classes:
            module = cls()
            with self.subTest(type=type(module)):
                self.assertEqual(len(module), 0)
                self.assertEqual(module.get("c", 42), 42)
                self.assertRaises(KeyError, module.__getitem__, "c")

    def test_list_creation(self):
        class DictLike(abc.Mapping):
            def __init__(self):
                self.list = ["a", "b", "a"]

            def __getitem__(self, key):
                return 42

            def __iter__(self):
                return iter(self.list)

            def __len__(self):
                return len(self.list)

        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                self.assertEqual(len(module), 3)
                self.assertEqual(module.__getitem__("a"), 1)
                self.assertEqual(module.__getitem__("b"), 2)
                self.assertListEqual(module.getall("a"), [1, 3])
                self.assertRaises(KeyError, module.__getitem__, "c")
                self.assertEqual(module.get("c", 42), 42)

                self.assertRaises(TypeError, cls, [], [])

                fromdict = cls(DictLike())
                self.assertEqual(len(fromdict), 3)
                self.assertEqual(fromdict.__getitem__("a"), 42)
                self.assertEqual(fromdict.__getitem__("b"), 42)
                self.assertListEqual(fromdict.getall("a"), [42, 42])
                self.assertRaises(KeyError, fromdict.__getitem__, "c")

    def test_dict_creation(self):
        for cls in self.classes:
            module = cls({"a": 1, "b": 2})
            with self.subTest(type=type(module)):
                self.assertEqual(len(module), 2)
                self.assertEqual(module.__getitem__("a"), 1)
                self.assertEqual(module.__getitem__("b"), 2)
                self.assertRaises(KeyError, module.__getitem__, "c")
                self.assertEqual(module.get("c", 42), 42)

    def test_keyword_creation(self):
        for cls in self.classes:
            module = cls(a=1, b=2)
            with self.subTest(type=type(module)):
                self.assertEqual(len(module), 2)
                self.assertEqual(module.__getitem__("a"), 1)
                self.assertEqual(module.__getitem__("b"), 2)
                self.assertRaises(KeyError, module.__getitem__, "c")
                self.assertEqual(module.get("c", 42), 42)

    def test_key_access(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                self.assertEqual(module.__getitem__("a"), 1)
                self.assertEqual(module.__getitem__("b"), 2)
                self.assertRaises(KeyError, module.__getitem__, "c")

    def test_index_access(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                self.assertEqual(module.__getitem__(0), ("a", 1))
                self.assertEqual(module.__getitem__(1), ("b", 2))
                self.assertEqual(module.__getitem__(2), ("a", 3))
                self.assertRaises(IndexError, module.__getitem__, 3)

    def test_slice_access(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                self.assertListEqual(
                    module.__getitem__(slice(0, 3)),
                    [("a", 1), ("b", 2), ("a", 3)],
                )
                self.assertListEqual(
                    module.__getitem__(slice(1, None)), [("b", 2), ("a", 3)]
                )
                self.assertListEqual(
                    module.__getitem__(slice(None, -1)), [("a", 1), ("b", 2)]
                )

    def test_set(self):
        for cls in self.classes:
            module = cls()
            with self.subTest(type=type(module)):
                module["a"] = 1
                module["b"] = 2
                module["a"] = 3

                self.assertEqual(module["a"], 3)
                self.assertEqual(module["b"], 2)
                self.assertListEqual(module.getall("a"), [3])
                self.assertEqual(len(module), 2)

                self.assertRaises(KeyError, module.__getitem__, "c")

                self.assertEqual(module.get("c", 42), 42)

    def test_delete(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                del module["a"]
                self.assertEqual(len(module), 1)
                self.assertRaises(KeyError, module.__getitem__, "a")
                self.assertRaises(KeyError, module.__getitem__, "c")

    def test_clear(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                module.clear()
                self.assertEqual(len(module), 0)
                self.assertRaises(KeyError, module.__getitem__, "a")
                self.assertRaises(KeyError, module.__getitem__, "b")
                self.assertRaises(KeyError, module.getall, "a")

                module["a"] = 42
                self.assertEqual(len(module), 1)
                self.assertEqual(module.__getitem__("a"), 42)

    def test_pop_noarg(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                self.assertTupleEqual(module.pop(), ("a", 3))
                self.assertEqual(len(module), 2)

    def test_update(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                module.update({"a": 42, "c": 7})
                self.assertEqual(len(module), 3)
                self.assertEqual(module.__getitem__("a"), 42)
                self.assertEqual(module.__getitem__("b"), 2)
                self.assertEqual(module.__getitem__("c"), 7)

                module.update()
                self.assertEqual(len(module), 3)
                self.assertEqual(module.__getitem__("a"), 42)
                self.assertEqual(module.__getitem__("b"), 2)
                self.assertEqual(module.__getitem__("c"), 7)

    def test_append(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                module.append("a", 42)
                self.assertEqual(len(module), 4)
                self.assertEqual(module.__getitem__("a"), 1)
                self.assertListEqual(module.getall("a"), [1, 3, 42])

                module.append("c", 43)
                self.assertEqual(len(module), 5)
                self.assertEqual(module.__getitem__("c"), 43)
                self.assertListEqual(module.getall("c"), [43])

    def test_len(self):
        for cls in self.classes:
            module = cls()
            with self.subTest(type=type(module)):
                self.assertEqual(len(module), 0)

                module = cls([("a", 1), ("b", 2), ("a", 3)])
                self.assertEqual(len(module), 3)

    def test_iterators(self):
        for cls in self.classes:
            module = cls()
            with self.subTest(type=type(module)):
                self.assertListEqual(list(module.items()), [])
                self.assertEqual(len(module.items()), 0)
                self.assertNotIn(("a", 1), module.items())

                self.assertListEqual(list(module.keys()), [])
                self.assertEqual(len(module.keys()), 0)
                self.assertNotIn(("a"), module.keys())

                self.assertListEqual(list(module.values()), [])
                self.assertEqual(len(module.values()), 0)
                self.assertNotIn(("1"), module.values())

                the_list = [("a", 1), ("b", 2), ("a", 3)]
                module = cls(the_list)

                self.assertListEqual(list(module.items()), the_list)
                self.assertEqual(len(module.items()), 3)
                self.assertIn(("a", 1), module.items())
                self.assertIn(("b", 2), module.items())
                self.assertIn(("a", 3), module.items())
                self.assertNotIn(("c", 4), module.items())

                self.assertListEqual(list(module.keys()), ["a", "b", "a"])
                self.assertEqual(len(module.keys()), 3)
                self.assertIn("a", module.keys())
                self.assertIn("b", module.keys())
                self.assertNotIn("c", module.keys())

                self.assertListEqual(list(module.values()), [1, 2, 3])
                self.assertEqual(len(module.values()), 3)
                self.assertIn(1, module.values())
                self.assertIn(2, module.values())
                self.assertIn(3, module.values())
                self.assertNotIn(4, module.values())

    def test_copy(self):
        for cls in self.classes:
            module = cls()
            with self.subTest(type=type(module)):
                copy = module.copy()
                self.assertEqual(module, copy)
                self.assertIsNot(module, copy)

                module["c"] = 42
                self.assertNotEqual(module, copy)

                module = cls([("a", 1), ("b", 2), ("a", 3)])
                copy = module.copy()
                self.assertEqual(module, copy)
                self.assertIsNot(module, copy)

                module["c"] = 42
                self.assertNotEqual(module, copy)

    def test_equality(self):
        classes = [
            (
                pvl.collections.PVLModule,
                pvl.collections.PVLGroup,
                pvl.collections.PVLObject,
            )
        ]
        try:
            from pvl.collections import PVLMultiDict

            classes.append(
                (
                    pvl.collections.PVLModuleNew,
                    pvl.collections.PVLGroupNew,
                    pvl.collections.PVLObjectNew,
                )
            )
        except ImportError:
            pass

        for modcls, grpcls, objcls in classes:
            module = modcls()
            group = grpcls()
            obj = objcls()
            with self.subTest(type=type(module)):
                self.assertFalse(module)
                self.assertFalse(group)
                self.assertFalse(obj)

                self.assertTrue(modcls(a=1))
                self.assertTrue(grpcls(a=1))
                self.assertTrue(objcls(a=1))

                self.assertNotEqual(modcls(), modcls(a=1))
                self.assertEqual(modcls(a=1), modcls(a=1))
                self.assertEqual(modcls(a=1), modcls([("a", 1)]))
                self.assertEqual(modcls(a=1), modcls({"a": 1}))
                self.assertNotEqual(modcls(a=1), modcls(b=1))
                self.assertNotEqual(modcls(a=1), modcls(a=2))

                self.assertNotIsInstance(group, modcls)
                self.assertNotIsInstance(group, objcls)

    def test_insert(self):
        the_list = [("a", 1), ("b", 2), ("a", 3)]
        for cls in self.classes:
            module = cls()
            with self.subTest(type=type(module)):
                self.assertRaises(TypeError, module.insert, "a")
                self.assertRaises(TypeError, module.insert, 0)
                module.insert(25, "key", "value")
                self.assertEqual(module, cls(key="value"))

                new_list = [("c", 4)] + the_list
                module = cls(the_list)
                module.insert(0, "c", 4)
                self.assertEqual(module, cls(new_list))

                module = cls(the_list)
                module.insert(0, ("c", 4))
                self.assertEqual(module, cls(new_list))

                module = cls(the_list)
                module.insert(0, {"c": 4})
                self.assertEqual(module, cls(new_list))

                module = cls(the_list)
                module.insert(0, [("c", 4)])
                self.assertEqual(module, cls(new_list))

                listinlist = list(the_list)
                listinlist.insert(1, ("c", 4))
                listinlist.insert(2, ("d", 5))
                module = cls(the_list)
                module.insert(1, [("c", 4), ("d", 5)])
                self.assertEqual(module, cls(listinlist))

    def test_key_index(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                self.assertRaises(KeyError, module.key_index, "error_key")
                self.assertRaises(IndexError, module.key_index, "a", 2)
                self.assertEqual(module.key_index("a"), 0)
                self.assertEqual(module.key_index("a", 0), 0)
                self.assertEqual(module.key_index("b"), 1)
                self.assertEqual(module.key_index("a", 1), 2)

    def test_insert_before(self):
        the_list = [("a", 1), ("b", 2), ("a", 3), ("c", 5)]
        expected = (
            (
                [("a", 4), ("a", 1), ("b", 2), ("a", 3), ("c", 5),],
                "a",
                0,
                [4, 1, 3],
                4,
            ),
            (
                [("a", 1), ("a", 4), ("b", 2), ("a", 3), ("c", 5),],
                "b",
                0,
                [1, 4, 3],
                1,
            ),
            (
                [("a", 1), ("b", 2), ("a", 4), ("a", 3), ("c", 5),],
                "a",
                1,
                [1, 4, 3],
                1,
            ),
            (
                [("a", 1), ("b", 2), ("a", 3), ("a", 4), ("c", 5),],
                "c",
                0,
                [1, 3, 4],
                1,
            ),
        )
        for cls in self.classes:
            module = cls(the_list)
            with self.subTest(type=type(module)):
                for (
                    expected_label,
                    key,
                    instance,
                    expected_list,
                    expected_value,
                ) in expected:
                    module = cls(the_list)
                    with self.subTest(
                        module=module,
                        expected_label=expected_label,
                        key=key,
                        instance=instance,
                        expected_list=expected_list,
                        expected_value=expected_value,
                    ):
                        exp_mod = cls(expected_label)
                        module.insert_before(key, [("a", 4)], instance)
                        self.assertEqual(exp_mod, module)
                        self.assertEqual(module["a"], expected_value)
                        self.assertListEqual(module.getall("a"), expected_list)

    def test_insert_after(self):
        the_list = [("a", 1), ("b", 2), ("a", 3), ("c", 5)]
        expected = (
            (
                [("a", 1), ("a", 4), ("b", 2), ("a", 3), ("c", 5),],
                "a",
                0,
                [1, 4, 3],
                1,
            ),
            (
                [("a", 1), ("b", 2), ("a", 4), ("a", 3), ("c", 5),],
                "b",
                0,
                [1, 4, 3],
                1,
            ),
            (
                [("a", 1), ("b", 2), ("a", 3), ("a", 4), ("c", 5),],
                "a",
                1,
                [1, 3, 4],
                1,
            ),
            (
                [("a", 1), ("b", 2), ("a", 3), ("c", 5), ("a", 4),],
                "c",
                0,
                [1, 3, 4],
                1,
            ),
        )
        for cls in self.classes:
            module = cls(the_list)
            with self.subTest(type=type(module)):
                for (
                    expected_label,
                    key,
                    instance,
                    expected_list,
                    expected_value,
                ) in expected:
                    module = cls(the_list)
                    with self.subTest(
                        expected_label=expected_label,
                        key=key,
                        instance=instance,
                        expected_list=expected_list,
                        expected_value=expected_value,
                    ):
                        exp_mod = cls(expected_label)
                        module.insert_after(key, [("a", 4)], instance)
                        self.assertEqual(exp_mod, module)
                        self.assertEqual(module["a"], expected_value)
                        self.assertListEqual(module.getall("a"), expected_list)

    def test_insert_before_after_raises(self):
        for cls in self.classes:
            module = cls([("a", 1), ("b", 2), ("a", 3)])
            with self.subTest(type=type(module)):
                self.assertRaises(
                    KeyError, module.insert_before, "error_key", [("fo", "ba")]
                )
                self.assertRaises(
                    KeyError, module.insert_after, "error_key", [("fo", "ba")]
                )
                self.assertRaises(
                    TypeError, module.insert_before, "a", [("fo", "ba"), 2]
                )
                self.assertRaises(
                    TypeError, module.insert_after, "a", [("fo", "ba"), 2]
                )


class TestDifferences(unittest.TestCase):
    def test_as_list(self):
        the_list = [("a", 1), ("b", 2)]

        # Returns list of tuples:
        old = OrderedMultiDict(the_list)
        self.assertListEqual(list(old), [("a", 1), ("b", 2)])

        try:
            from pvl.collections import PVLMultiDict

            # Returns list of keys, which is semantically identical to calling
            # list() on a dict.
            new = PVLMultiDict(the_list)
            self.assertListEqual(list(new), ["a", "b"])
        except ImportError:
            pass

    def test_discard(self):
        the_list = [("a", 1), ("b", 2), ("a", 3)]

        # Has a set-like .discard() function
        old = OrderedMultiDict(the_list)
        old.discard("a")
        self.assertEqual(len(old), 1)
        self.assertRaises(KeyError, old.getall, "a")
        self.assertRaises(KeyError, old.__getitem__, "a")

        self.assertEqual(old.__getitem__("b"), 2)
        old.discard("b")
        self.assertEqual(len(old), 0)
        self.assertRaises(KeyError, old.__getitem__, "b")

        old.discard("c")
        self.assertEqual(len(old), 0)

        try:
            from pvl.collections import PVLMultiDict

            # Does not have a set-like .discard() function,
            # because it isn't a set!
            new = PVLMultiDict(the_list)
            self.assertRaises(AttributeError, getattr, new, "discard")
        except ImportError:
            pass

    def test_pop(self):
        the_list = [("a", 1), ("b", 2), ("a", 3)]

        # Removes all keys that match, but returns only the first value, which
        # is weird
        old = OrderedMultiDict(the_list)
        self.assertEqual(old.pop("a"), 1)
        self.assertEqual(len(old), 1)
        self.assertRaises(KeyError, old.getall, "a")
        self.assertRaises(KeyError, old.pop, "a")
        self.assertEqual(old.pop("a", 42), 42)

        self.assertEqual(old.pop("b"), 2)
        self.assertEqual(len(old), 0)
        self.assertRaises(KeyError, old.pop, "b")
        self.assertRaises(KeyError, old.__getitem__, "b")

        self.assertRaises(KeyError, old.pop, "c")
        self.assertEqual(old.pop("c", 42), 42)

        try:
            from pvl.collections import PVLMultiDict

            # Removes only the first key
            new = PVLMultiDict(the_list)
            self.assertEqual(new.pop("a"), 1)
            self.assertEqual(len(new), 2)
            self.assertListEqual(new.getall("a"), [3,])
            self.assertEqual(new.pop("a"), 3)
            self.assertEqual(new.pop("a", 42), 42)

            self.assertEqual(new.pop("b"), 2)
            self.assertEqual(len(new), 0)
            self.assertRaises(KeyError, new.pop, "b")
            self.assertRaises(KeyError, new.__getitem__, "b")

            self.assertRaises(KeyError, new.pop, "c")
            self.assertEqual(new.pop("c", 42), 42)
        except ImportError:
            pass

    def test_popitem(self):
        the_list = [("a", 1), ("b", 2), ("a", 3)]

        # Removes the last item
        old = OrderedMultiDict(the_list)
        self.assertTupleEqual(old.popitem(), ("a", 3))
        self.assertEqual(len(old), 2)
        self.assertTupleEqual(old.popitem(), ("b", 2))
        self.assertEqual(len(old), 1)
        self.assertTupleEqual(old.popitem(), ("a", 1))
        self.assertEqual(len(old), 0)
        self.assertRaises(KeyError, old.popitem)

        try:
            from pvl.collections import PVLMultiDict

            # Removes a random item, in proper dict-like fashion
            new = PVLMultiDict(the_list)
            self.assertIn(new.popitem(), the_list)
            self.assertEqual(len(new), 2)
            new.popitem()
            new.popitem()
            self.assertRaises(KeyError, new.popitem)
        except ImportError:
            pass

    def test_repr(self):
        # Original repr
        old = OrderedMultiDict()
        self.assertEqual(repr(old), "OrderedMultiDict([])")

        try:
            from pvl.collections import PVLMultiDict

            # MultiDict repr
            new = PVLMultiDict()
            self.assertEqual(repr(new), "PVLMultiDict()")
        except ImportError:
            pass

    def test_py3_items(self):
        the_list = [("a", 1), ("b", 2), ("a", 3)]

        # These views are returned as lists!
        old = OrderedMultiDict(the_list)
        self.assertIsInstance(old.items(), pvl.collections.ItemsView)
        self.assertIsInstance(old.keys(), pvl.collections.KeysView)
        self.assertIsInstance(old.values(), pvl.collections.ValuesView)
        views = [
            (old.items(), old.keys(), old.values()),
        ]

        try:
            from pvl.collections import PVLMultiDict

            # These are proper Python 3 views:
            new = PVLMultiDict(the_list)
            self.assertIsInstance(new.items(), abc.ItemsView)
            self.assertIsInstance(new.keys(), abc.KeysView)
            self.assertIsInstance(new.values(), abc.ValuesView)
            views.append(
                (list(new.items()), list(new.keys()), list(new.values()))
            )
        except ImportError:
            pass

        # However, if you wrap the new items in a list (as above), this is
        # the same:
        for items, keys, values in views:
            self.assertTupleEqual(items[0], ("a", 1))
            self.assertTupleEqual(items[1], ("b", 2))
            self.assertTupleEqual(items[2], ("a", 3))
            self.assertEqual(items.index(("a", 1)), 0)
            self.assertEqual(items.index(("b", 2)), 1)
            self.assertEqual(items.index(("a", 3)), 2)

            self.assertEqual(keys[0], "a")
            self.assertEqual(keys[1], "b")
            self.assertEqual(keys[2], "a")
            self.assertEqual(keys.index("a"), 0)
            self.assertEqual(keys.index("b"), 1)

            self.assertEqual(values[0], 1)
            self.assertEqual(values[1], 2)
            self.assertEqual(values[2], 3)
            self.assertEqual(values.index(1), 0)
            self.assertEqual(values.index(2), 1)
            self.assertEqual(values.index(3), 2)

    def test_conversion(self):
        the_list = [("a", 1), ("b", 2), ("a", 3)]

        # This returns a list of key, value tuple pairs
        old = OrderedMultiDict(the_list)
        self.assertListEqual(list(old), the_list)

        # === Callling dict(old) ===
        # This is the one test that I could not get to pass from
        # pvl 0.3, and now I know why: it is because the
        # OrderedMultiDict inherently carries two copies of the
        # information added to it: one in the default dict that
        # it inherits from being directly subclassed from dict,
        # and the other is an internal list of key, value tuples.
        #
        # When I tried to run this test using the regular Python
        # interpreters, I got
        # self.assertEqual(dict(old), {'a': 1, 'b': 2})
        # This is because when it is passed to the dict constructor,
        # Python makes it into a regular Mapping object with unique
        # keys, inevitably loosing the double value of 'a'.
        #
        # I could not understand why pvl 0.3 had this test that it
        # expected to pass:
        # expected_dict = {
        #     'a': [1, 3],
        #     'b': [2],
        # }
        #
        # That was until Travis ran the tests using both the regular
        # python interpreter and pypy (I don't usually run pypy tests locally).
        # If you run the test with a regular python interpreter, you get the
        # first result, with the second value of 'a' being lost.  If you run
        # the test with the pypy3 interpreter, you get the second result, with
        # the values for the keys being lists!
        #
        # This scared the heck out of me.
        #
        # This is what I think is happening:
        # When you run dict() on an OrderedMultiDict, the python interpreter
        # uses the fact that it is a Mapping item, and internally runs its
        # .items() function to get the (key, value) pairs and builds a new
        # dict from them (it encounters a second value for the key 'a', and
        # looses it).  The pypy3 interpreter, must not use the .items()
        # function on the passed-in Mapping object, but notices that this
        # object derives from a dict, and so just grabs the internal dict
        # representation, which OrderedMultiDict hides from the user, but
        # is implemented as single keys with lists as the values.
        #
        # And that's why the same code produces different results with
        # different interpreters.

        try:
            from pvl.collections import PVLMultiDict

            # This returns the same thing that calling list() on a dict would,
            # the list of keys
            new = PVLMultiDict(the_list)
            self.assertListEqual(list(new), ["a", "b", "a"])
            self.assertEqual(dict(new), {"a": 1, "b": 2})
        except ImportError:
            pass

    def test_equality(self):

        # There is an isinstance() check in the __eq__ operator
        oldmod = pvl.collections.PVLModule()
        oldgrp = pvl.collections.PVLGroup()
        oldobj = pvl.collections.PVLObject()

        self.assertEqual(oldmod, oldmod)
        self.assertNotEqual(oldmod, oldgrp)
        self.assertNotEqual(oldmod, oldobj)

        self.assertNotEqual(oldgrp, oldmod)
        self.assertEqual(oldgrp, oldgrp)
        self.assertNotEqual(oldgrp, oldobj)

        self.assertNotEqual(oldobj, oldmod)
        self.assertNotEqual(oldobj, oldgrp)
        self.assertEqual(oldobj, oldobj)

        try:
            from pvl.collections import PVLMultiDict

            # Value-based notion of equality
            newmod = pvl.collections.PVLModuleNew()
            newgrp = pvl.collections.PVLGroupNew()
            newobj = pvl.collections.PVLObjectNew()

            self.assertEqual(newmod, newgrp)
            self.assertEqual(newmod, newobj)
        except ImportError:
            pass



# -*- coding: utf-8 -*-
"""
Exceptions for the Parameter Value Library.
"""

# Copyright 2019-2020, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.


def firstpos(sub: str, pos: int):
    """On the assumption that *sub* is a substring contained in a longer
    string, and *pos* is the index in that longer string of the final
    character in sub, returns the position of the first character of
    sub in that longer string.

    This is useful in the PVL library when we know the position of the
    final character of a token, but want the position of the first
    character.
    """
    return pos - len(sub) + 1


def linecount(doc: str, end: int, start: int = 0):
    """Returns the number of lines (by counting the
    number of newline characters \\n, with the first line
    being line number one) in the string *doc* between the
    positions *start* and *end*.
    """
    return doc.count("\n", start, end) + 1


class LexerError(ValueError):
    """Subclass of ValueError with the following additional properties:

    msg: The unformatted error message
    doc: The PVL text being parsed
    pos: The start index in doc where parsing failed
    lineno: The line corresponding to pos
    colno: The column corresponding to pos
    """

    def __init__(self, msg, doc, pos, lexeme):
        self.pos = firstpos(lexeme, pos)
        lineno = linecount(doc, self.pos)
        colno = self.pos - doc.rfind("\n", 0, self.pos)
        # Assemble a context string that consists of whole
        # words, using fragments is hard to read.
        context_tokens = doc[self.pos - 15: self.pos + 15].split(" ")
        context = " ".join(context_tokens[1:-1])
        errmsg = (
            f"{msg}: line {lineno} column {colno} (char {pos}) "
            f'near "{context}"'
        )
        super().__init__(self, errmsg)
        self.msg = msg
        self.doc = doc
        self.lineno = lineno
        self.colno = colno
        self.lexeme = lexeme

    def __reduce__(self):
        return self.__class__, (self.msg, self.doc, self.pos, self.lexeme)


class ParseError(Exception):
    """An exception to signal errors in the pvl parser."""

    def __init__(self, msg, token=None):
        super().__init__(self, msg)
        self.token = token


class QuantityError(Exception):
    """A simple exception to distinguish errors from Quantity classes."""

    pass



#!/usr/bin/env python
"""This module has tests for the pvl grammar regexes."""

# Copyright 2019, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import unittest

from pvl.grammar import PVLGrammar


class TestLeapSeconds(unittest.TestCase):
    def setUp(self):
        self.g = PVLGrammar()

    def test_H_frag(self):
        for s in ("00", "05", "10", "19", "23"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._H_frag, s)
                self.assertEqual(s, m.groupdict()["hour"])

    def test_M_frag(self):
        for s in ("00", "05", "10", "19", "23", "59"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._M_frag, s)
                self.assertEqual(s, m.groupdict()["minute"])

    def test_f_frag(self):
        for s in (".1", ".123", ".789"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._f_frag, s)
                self.assertEqual(s[1:], m.groupdict()["microsecond"])
        for s in ("1", "", "abc"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._f_frag, s)
                self.assertEqual(None, m)

    def test_Y_frag(self):
        for s in ("0001", "1985", "9999"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._Y_frag, s)
                self.assertEqual(s, m.groupdict()["year"])
        for s in ("0000", "10000", "abc"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._Y_frag, s)
                self.assertEqual(None, m)

    def test_m_frag(self):
        for s in ("01", "04", "11", "12"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._m_frag, s)
                self.assertEqual(s, m.groupdict()["month"])
        for s in ("1", "00", "abc"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._m_frag, s)
                self.assertEqual(None, m)

    def test_d_frag(self):
        for s in ("01", "10", "28", "31"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._d_frag, s)
                self.assertEqual(s, m.groupdict()["day"])
        for s in ("1", "00", "32"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._d_frag, s)
                self.assertEqual(None, m)

    def test_j_frag(self):
        for s in ("001", "123", "234", "366"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._j_frag, s)
                self.assertEqual(s, m.groupdict()["doy"])
        for s in ("1", "30", "367"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._j_frag, s)
                self.assertEqual(None, m)

    def test_Ymd_frag(self):
        for s in ("2001-01-01", "2019-12-04"):
            with self.subTest(string=s):
                self.assertIsNotNone(re.fullmatch(self.g._Ymd_frag, s))
        for s in ("2001-1-1", "2019-12-00"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._Ymd_frag, s)
                self.assertEqual(None, m)

    def test_Yj_frag(self):
        for s in ("2001-001", "2019-180"):
            with self.subTest(string=s):
                self.assertIsNotNone(re.fullmatch(self.g._Yj_frag, s))
        for s in ("2001-1-1", "2019-367"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._Yj_frag, s)
                self.assertEqual(None, m)

    def test_time_frag(self):
        for s in ("01:02:60", "02:03:60Z", "03:04:60.123", "03:04:60.123Z"):
            with self.subTest(string=s):
                self.assertIsNotNone(re.fullmatch(self.g._time_frag, s))
        for s in ("01:02:59", "23:56"):
            with self.subTest(string=s):
                m = re.fullmatch(self.g._time_frag, s)
                self.assertEqual(None, m)

    def test_leap_second_Ymd_re(self):
        for s in ("2001-12-31T01:59:60.123Z", "01:59:60Z", "01:59:60"):
            with self.subTest(string=s):
                self.assertIsNotNone(self.g.leap_second_Ymd_re.fullmatch(s))

    def test_leap_second_Yj_re(self):
        for s in ("2001-180T01:59:60.123Z", "01:59:60Z", "01:59:60"):
            with self.subTest(string=s):
                self.assertIsNotNone(self.g.leap_second_Yj_re.fullmatch(s))

    def test_allowed(self):
        for c in ("a", "b", " ", "\n"):
            with self.subTest(char=c):
                self.assertTrue(self.g.char_allowed(c))

        for c in ("\b", chr(127)):
            with self.subTest(char=c):
                self.assertFalse(self.g.char_allowed(c))



# -*- coding: utf-8 -*-
"""Python implementation of PVL (Parameter Value Language), with upcoming
features.

If you currently use::

    import pvl

you can change to::

    import pvl.new as pvl

And then use all of the pvl functions as you usually would.  You
will also need to have the 3rd party multidict library
(https://github.com/aio-libs/multidict, conda installable) installed.
But then, any objects that are returned by the load functions will
be the new PVLMultiDict objects.
"""

# Copyright 2015, 2017, 2019-2020, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import inspect
import urllib.request

from pvl import *  # noqa: F401,F403
from pvl import get_text_from, decode_by_char

from .parser import PVLParser, OmniParser
from .collections import PVLModuleNew, PVLGroupNew, PVLObjectNew

__all__ = [
    "PVLModuleNew",
    "PVLGroupNew",
    "PVLObjectNew",
]


def load(path, parser=None, grammar=None, decoder=None, **kwargs):
    """Returns a Python object from parsing the file at *path*.

    :param path: an :class:`os.PathLike` which presumably has a
        PVL Module in it to parse.
    :param parser: defaults to :class:`pvl.parser.OmniParser()`.
    :param grammar: defaults to :class:`pvl.grammar.OmniGrammar()`.
    :param decoder: defaults to :class:`pvl.decoder.OmniDecoder()`.
    :param ``**kwargs``: the keyword arguments that will be passed
        to :func:`loads()` and are described there.

    If *path* is not an :class:`os.PathLike`, it will be assumed to be an
    already-opened file object, and ``.read()`` will be applied
    to extract the text.

    If the :class:`os.PathLike` or file object contains some bytes
    decodable as text, followed by some that is not (e.g. an ISIS
    cube file), that's fine, this function will just extract the
    decodable text.
    """
    return loads(
        get_text_from(path),
        parser=parser,
        grammar=grammar,
        decoder=decoder,
        **kwargs
    )


def loadu(url, parser=None, grammar=None, decoder=None, **kwargs):
    """Returns a Python object from parsing *url*.

        :param url: this will be passed to :func:`urllib.request.urlopen`
            and can be a string or a :class:`urllib.request.Request` object.
        :param parser: defaults to :class:`pvl.parser.OmniParser()`.
        :param grammar: defaults to :class:`pvl.grammar.OmniGrammar()`.
        :param decoder: defaults to :class:`pvl.decoder.OmniDecoder()`.
        :param ``**kwargs``: the keyword arguments that will be passed
            to :func:`urllib.request.urlopen` and to :func:`loads()`.

        The ``**kwargs`` will first be scanned for arguments that
        can be given to :func:`urllib.request.urlopen`.  If any are
        found, they are extracted and used.  All remaining elements
        will be passed on as keyword arguments to :func:`loads()`.

        Note that *url* can be any URL that :func:`urllib.request.urlopen`
        takes.  Certainly http and https URLs, but also file, ftp, rsync,
        sftp and more!
        """

    # Peel off the args for urlopen:
    url_args = dict()
    for a in inspect.signature(urllib.request.urlopen).parameters.keys():
        if a in kwargs:
            url_args[a] = kwargs.pop(a)

    # The object returned from urlopen will always have a .read()
    # function that returns bytes, so:
    with urllib.request.urlopen(url, **url_args) as resp:
        s = decode_by_char(resp)

    return loads(s, parser=parser, grammar=grammar, decoder=decoder, **kwargs)


def loads(s: str, parser=None, grammar=None, decoder=None, **kwargs):
    """Deserialize the string, *s*, as a Python object.

    :param s: contains some PVL to parse.
    :param parser: defaults to :class:`pvl.parser.OmniParser() which will
        return the new PVLMultiDict-derived objects`.
    :param grammar: defaults to :class:`pvl.grammar.OmniGrammar()`.
    :param decoder: defaults to :class:`pvl.decoder.OmniDecoder()`.
    :param ``**kwargs``: the keyword arguments to pass to the *parser* class
        if *parser* is none.
    """
    if isinstance(s, bytes):
        # Someone passed us an old-style bytes sequence.  Although it isn't
        # a string, we can deal with it:
        s = s.decode()

    if parser is None:
        parser = OmniParser(
            grammar=grammar,
            decoder=decoder,
            module_class=PVLModuleNew,
            group_class=PVLGroupNew,
            object_class=PVLObjectNew,
            **kwargs
        )
    elif not isinstance(parser, PVLParser):
        raise TypeError("The parser must be an instance of pvl.PVLParser.")

    return parser.parse(s)



# -*- coding: utf-8 -*-

import datetime
import unittest

from pvl.encoder import PVLEncoder, ODLEncoder, PDSLabelEncoder
from pvl.collections import Quantity, PVLModule, PVLGroup, PVLObject


class TestDecoder(unittest.TestCase):
    def setUp(self):
        self.e = PVLEncoder()

    def test_format(self):
        s = "ABC"
        self.assertEqual(s, self.e.format(s))
        self.assertEqual(f"  {s}", self.e.format(s, 1))

        s = "keyword = " + ("a" * 100)
        self.assertEqual(s, self.e.format(s))

        k = "keyword = "
        a60 = "a" * 60
        b60 = "b" * 60
        s = k + a60 + " " + b60
        self.assertEqual(
            (k + a60 + "\n" + (" " * len(k)) + b60), self.e.format(s)
        )

    def test_encode_string(self):
        s = "ABC"
        self.assertEqual(s, self.e.encode_string(s))

        s = 'AB"CD'
        self.assertEqual(f"'{s}'", self.e.encode_string(s))

        s = """AB'CD"""
        self.assertEqual(f'''"{s}"''', self.e.encode_string(s))

        s = "12:01"
        self.assertEqual(f'"{s}"', self.e.encode_string(s))

        s = "AB CD"
        self.assertEqual(f'"{s}"', self.e.encode_string(s))

        s = """Both"kinds'of quotes"""
        self.assertRaises(ValueError, self.e.encode_string, s)

    def test_encode_date(self):
        t = datetime.date(2019, 12, 31)
        self.assertEqual("2019-12-31", self.e.encode_date(t))

    def test_encode_time(self):
        t = datetime.time(1, 2)
        self.assertEqual("01:02", self.e.encode_time(t))

        t = datetime.time(13, 14, 15)
        self.assertEqual("13:14:15", self.e.encode_time(t))

        t = datetime.time(23, 24, 25, 123)
        self.assertEqual("23:24:25.000123", self.e.encode_time(t))

        t = datetime.time(23, 24, 0, 123)
        self.assertEqual("23:24:00.000123", self.e.encode_time(t))

    def test_encode_datetime(self):
        t = datetime.datetime(2001, 1, 1, 2, 3)
        self.assertEqual("2001-01-01T02:03", self.e.encode_datetime(t))

    def test_encode_set(self):
        f = frozenset(["a", "b", "c"])
        s_list = self.e.encode_set(f).strip("{}").split(", ")
        self.assertEqual(f, set(s_list))

    def test_encode_sequence(self):
        s = ["a", "b", "c"]
        self.assertEqual("(a, b, c)", self.e.encode_sequence(s))

    def test_encode_simple_value(self):
        pairs = (
            (None, "NULL"),
            (["a", "b", "c"], "(a, b, c)"),
            (datetime.datetime(2001, 1, 1, 2, 3), "2001-01-01T02:03"),
            (True, "TRUE"),
            (1.23, "1.23"),
            (42, "42"),
            ("ABC", "ABC"),
        )
        for p in pairs:
            with self.subTest(pair=p):
                self.assertEqual(p[1], self.e.encode_simple_value(p[0]))

    def test_encode_value(self):
        pairs = ((42, "42"), (Quantity(34, "m/s"), "34 <m/s>"))
        for p in pairs:
            with self.subTest(pair=p):
                self.assertEqual(p[1], self.e.encode_value(p[0]))

    def test_encode_assignment(self):
        self.assertEqual("a = b;", self.e.encode_assignment("a", "b"))

    def test_encode_aggregation_block(self):
        g = PVLGroup(a="b", c="d")
        s = "BEGIN_GROUP = foo;\n  a = b;\n  c = d;\nEND_GROUP = foo;"
        self.assertEqual(s, self.e.encode_aggregation_block("foo", g))

    def test_encode_module(self):
        m = PVLModule(foo=PVLGroup(a="b", c="d"))
        s = "BEGIN_GROUP = foo;\n  a = b;\n  c = d;\nEND_GROUP = foo;"
        self.assertEqual(s, self.e.encode_module(m))

    def test_encode(self):
        g = PVLGroup(a="b", c="d")
        s = "a = b;\nc = d;\nEND;"
        self.assertEqual(s, self.e.encode(g))

        m = PVLModule(foo=PVLGroup(a="b", c="d"))
        s = "BEGIN_GROUP = foo;\n  a = b;\n  c = d;\nEND_GROUP = foo;\nEND;"
        self.assertEqual(s, self.e.encode(m))

        m = PVLModule(
            foo=PVLGroup(
                a="b", c="d", newline="Should be quoted\nand two lines."
            )
        )
        s = """BEGIN_GROUP = foo;
  a       = b;
  c       = d;
  newline = "Should be quoted
and two lines.";
END_GROUP = foo;
END;"""
        self.assertEqual(s, self.e.encode(m))

    def test_encode_quantity(self):
        q, s = Quantity(34, "m/s"), "34 <m/s>"
        self.assertEqual(s, self.e.encode_quantity(q))

        self.assertRaises(ValueError, self.e.encode_quantity, "not a quant")

        try:
            from astropy import units as u

            # astropy.units.Quantity makes values floating point
            # and appears to print the units with spaces.
            # It doesn't violate the PVL spec, but may cause users
            # a surprise.  However, if they are using Astropy
            # Quantities, they are theoretically aware of all that.
            q, s = u.Quantity(34, "m/s"), "34.0 <m / s>"
            self.assertEqual(s, self.e.encode_quantity(q))
        except ImportError:  # astropy isn't available.
            pass

        try:
            from pint import Quantity as pintquant

            # pint.Quantity also has its own peculiarities about
            # formating the output.  Again, doesn't break the spec,
            # but may cause surprises.
            q, s = pintquant(34, "m/s"), "34 <meter / second>"
            self.assertEqual(s, self.e.encode_quantity(q))
        except ImportError:  # pint isn't available.
            pass


class TestODLDecoder(unittest.TestCase):
    def setUp(self):
        self.e = ODLEncoder()

    def test_is_scalar(self):
        self.assertTrue(self.e.is_scalar(5))
        self.assertTrue(self.e.is_scalar("scalar"))
        self.assertTrue(self.e.is_scalar(Quantity(5, "m")))
        self.assertFalse(self.e.is_scalar(Quantity("five", "m")))

    def test_encode_quantity(self):
        try:
            from astropy import units as u

            p = (u.Quantity(34, "m/s"), "34.0 <m / s>")
            self.assertEqual(p[1], self.e.encode_quantity(p[0]))
        except ImportError:  # astropy isn't available.
            pass


class TestPDSLabelEncoder(unittest.TestCase):
    def setUp(self):
        self.e = PDSLabelEncoder()

    def test_count_aggs(self):
        m = PVLModule(a=PVLGroup(), b=PVLObject(), c=PVLObject())
        self.assertEqual((2, 1), self.e.count_aggs(m))

    def test_is_PDSgroup(self):
        g = PVLGroup(a="b", c=PVLGroup())
        self.assertFalse(self.e.is_PDSgroup(g))

        g = PVLGroup({"a": "b", "^c": 5})
        self.assertFalse(self.e.is_PDSgroup(g))

        g = PVLGroup({"a": "b", "^c": "d"})
        self.assertTrue(self.e.is_PDSgroup(g))

        g = PVLGroup((("a", "b"), ("c", "d"), ("a", "b2")))
        self.assertFalse(self.e.is_PDSgroup(g))

    def test_convert_grp_to_obj(self):
        g = PVLGroup(a="b", c="d")
        o = PVLObject(a="b", c="d")
        converted = PVLObject(g)
        self.assertEqual(o, converted)
        self.assertIsInstance(converted, PVLObject)

    def test_encode_aggregation_block(self):
        g = PVLGroup(a="b", c=PVLGroup(d="e"))
        no_convert = PDSLabelEncoder(convert_group_to_object=False)
        self.assertRaises(
            ValueError, no_convert.encode_aggregation_block, "key", g
        )

        s = """OBJECT = key\r
  A = b\r
  GROUP = c\r
    D = e\r
  END_GROUP = c\r
END_OBJECT = key"""
        self.assertEqual(s, self.e.encode_aggregation_block("key", g))

    def test_encode_set(self):
        s = {"a", "has'apostrophe"}
        self.assertRaises(ValueError, self.e.encode_set, s)

        s = {"a", "has\newline"}
        self.assertRaises(ValueError, self.e.encode_set, s)

    def test_encode_time(self):
        t = datetime.time(1, 2)
        self.assertEqual("01:02Z", self.e.encode_time(t))

        t = datetime.time(13, 14, 15,)
        self.assertEqual("13:14:15Z", self.e.encode_time(t))

        t = datetime.time(
            13, 14, 15, tzinfo=datetime.timezone(datetime.timedelta(hours=2))
        )
        self.assertRaises(ValueError, self.e.encode_time, t)

        t = datetime.time(
            13, 14, 15, tzinfo=datetime.timezone(datetime.timedelta(hours=0))
        )
        self.assertRaises(ValueError, self.e.encode_time, t)

    def test_encode(self):
        m = PVLModule(a=PVLGroup(g1=2, g2=3.4), b="c")

        no_convert = PDSLabelEncoder(convert_group_to_object=False)
        self.assertRaises(ValueError, no_convert.encode, m)

        s = """OBJECT = a\r
  G1 = 2\r
  G2 = 3.4\r
END_OBJECT = a\r
B = c\r
END\r\n"""
        self.assertEqual(s, self.e.encode(m))

        m = PVLModule(
            a="b",
            staygroup=PVLGroup(c="d"),
            obj=PVLGroup(d="e", f=PVLGroup(g="h")),
        )

        s = """A = b\r
GROUP = staygroup\r
  C = d\r
END_GROUP = staygroup\r
OBJECT = obj\r
  D = e\r
  GROUP = f\r
    G = h\r
  END_GROUP = f\r
END_OBJECT = obj\r
END\r\n"""

        self.assertEqual(s, self.e.encode(m))



#!/usr/bin/env python
"""This module has tests for the pvl lang functions."""

# Copyright 2019, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pvl.grammar import PVLGrammar
from pvl.token import Token


class TestToken(unittest.TestCase):
    def test_init(self):
        s = "token"
        self.assertEqual(s, Token(s))
        self.assertEqual(s, Token(s, grammar=PVLGrammar()))

    def test_is_comment(self):
        c = Token("/* comment */")
        self.assertTrue(c.is_comment())

        n = Token("not comment */")
        self.assertFalse(n.is_comment())

    def test_is_begin_aggregation(self):
        for s in ("BEGIN_GROUP", "Begin_Group", "ObJeCt"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_begin_aggregation())

        b = Token("END_GROUP")
        self.assertFalse(b.is_begin_aggregation())

    def test_is_end_statement(self):
        t = Token("END")
        self.assertTrue(t.is_end_statement())

        t = Token("Start")
        self.assertFalse(t.is_end_statement())

    def test_is_datetime(self):
        for s in (
            "2001-027T23:45",
            "2001-01-01T01:34Z",
            "01:42:57Z",
            "23:45",
            "01:42:57",
            "12:34:56.789",
            "2001-01-01",
            "2001-027",
        ):
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_datetime())

        for s in ("3:450", "frank"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_datetime())

    def test_is_parameter_name(self):
        for s in ("Hello", "ProductId"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_parameter_name())

        for s in ("Group", "/*comment*/", "2001-027"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_parameter_name())

    def test_is_decimal(self):
        for s in (
            "125",
            "+211109",
            "-79",  # Integers
            "69.35",
            "+12456.345",
            "-0.23456",
            ".05",
            "-7.",  # Floating
            "-2.345678E12",
            "1.567E-10",
            "+4.99E+3",
        ):  # Exponential
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_decimal())

        for s in ("2#0101#", "frank"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_decimal())

    def test_is_binary(self):
        for s in ("2#0101#", "+2#0101#", "-2#0101#"):
            with self.subTest(string=s):
                t = Token(s)
                # self.assertTrue(t.is_binary())
                self.assertTrue(t.is_non_decimal())

        # for s in ('+211109', 'echo', '+8#0156#'):
        for s in ("+211109", "echo"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_non_decimal())

    def test_is_octal(self):
        for s in ("8#0107#", "+8#0156#", "-8#0134#"):
            with self.subTest(string=s):
                t = Token(s)
                # self.assertTrue(t.is_octal())
                self.assertTrue(t.is_non_decimal())

        # for s in ('+211109', 'echo', '2#0101#'):
        #     with self.subTest(string=s):
        #         t = Token(s)
        #         self.assertFalse(t.is_octal())

    def test_is_hex(self):
        for s in ("16#100A#", "+16#23Bc#", "-16#98ef#"):
            with self.subTest(string=s):
                t = Token(s)
                # self.assertTrue(t.is_hex())
                self.assertTrue(t.is_non_decimal())

        # for s in ('+211109', 'echo', '2#0101#', '8#0107#'):
        #     with self.subTest(string=s):
        #         t = Token(s)
        #         self.assertFalse(t.is_hex())

    def test_isnumeric(self):
        for s in (
            "125",
            "+211109",
            "-79",  # Integers
            "69.35",
            "+12456.345",
            "-0.23456",
            ".05",
            "-7.",  # Floating
            "-2.345678E12",
            "1.567E-10",
            "+4.99E+3",  # Exponential
            "2#0101#",
            "+2#0101#",
            "-2#0101#",  # Binary
            "8#0107#",
            "+8#0156#",
            "-8#0134#",  # Octal
            "16#100A#",
            "+16#23Bc#",
            "-16#98ef#",
        ):  # Hex
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.isnumeric())

        for s in ("frank", "#", "-apple"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.isnumeric())

    def test_is_space(self):
        for s in ("  ", "\t\n"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_space())

        for s in ("not space", ""):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_space())

    def test_is_WSC(self):
        for s in (" /*com*/  ", "/*c1*/\n/*c2*/"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_WSC())

        for s in (" /*com*/ not comment", " surrounding "):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_WSC())

    def test_is_delimiter(self):
        t = Token(";")
        self.assertTrue(t.is_delimiter())

        t = Token("not")
        self.assertFalse(t.is_delimiter())

    def test_is_unquoted_string(self):
        for s in ("Hello", "Product", "Group"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_unquoted_string())

        for s in ("/*comment*/", "2001-027", '"quoted"'):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_unquoted_string())

    def test_is_quoted_string(self):
        for s in ('"Hello &"', "'Product Id'", '""'):
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_quoted_string())

        for s in ("/*comment*/", "2001-027", '"'):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_quoted_string())

    def test_is_string(self):
        for s in (
            '"Hello &"',
            "'Product Id'",
            '""',
            "Hello",
            "Product",
            "Group",
        ):
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_string())

        for s in ("/*comment*/", "2001-027", '"'):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_string())

    def test_is_simple_value(self):
        for s in (
            '"Hello &"',
            "'Product Id'",
            '""',  # Quoted Strings
            "Hello",
            "Group" "2001-01-01",  # Unquoted Strings
            "2001-027",  # Date
            "23:45",
            "01:42:57",
            "12:34:56.789" "2001-027T23:45",  # Time
            "2001-01-01T01:34Z",  # Datetime
            "125",
            "+211109",
            "-79",  # Integers
            "69.35",
            "+12456.345",
            "-0.23456",
            ".05",
            "-7.",  # Floating
            "-2.345678E12",
            "1.567E-10",
            "+4.99E+3",  # Exponential
            "2#0101#",
            "+2#0101#",
            "-2#0101#",  # Binary
            "8#0107#",
            "+8#0156#",
            "-8#0134#",  # Octal
            "16#100A#",
            "+16#23Bc#",
            "-16#98ef#",
        ):  # Hex
            with self.subTest(string=s):
                t = Token(s)
                self.assertTrue(t.is_simple_value())

        for s in ("/*comment*/", "=", '"', "{", "(", "Product Id"):
            with self.subTest(string=s):
                t = Token(s)
                self.assertFalse(t.is_simple_value())

    def test_split(self):
        s = "Hello Bob"
        t = Token(s)
        t_list = t.split()
        for x in t_list:
            with self.subTest(token=x):
                self.assertIsInstance(x, Token)



# -*- coding: utf-8 -*-
"""Parameter Value Language parser.

The definition of PVL used in this module is based on the Consultive
Committee for Space Data Systems, and their Parameter Value
Language Specification (CCSD0006 and CCSD0008), CCSDS 6441.0-B-2,
referred to as the Blue Book with a date of June 2000.

Some of the documention in this module represents the structure
diagrams from the Blue Book for parsing PVL in a Backus–Naur
form.

So Figure 1-1 from the Blue Book would be represented as :

 <Item-A> ::= ( [ <Item-B>+ | <Item-C> ] <Item-D> )*

Finally, the Blue Book defines <WSC> as a possibly empty collection
of white space characters or comments:

 <WSC> ::= ( <white-space-character> | <comment> )*

However, to help remember that <WSC> could be empty, we will typically
always show it as <WSC>*.

Likewise the <Statement-Delimiter> is defined as:

 <Statement-Delimiter> ::= <WSC>* [ ';' | <EOF> ]

However, since all elements are optional, we will typically
show it as [<Statement-Delimiter>].

The parser deals with managing the tokens that come out of the lexer.
Once the parser gets to a state where it has something that needs to
be converted to a Python object and returned, it uses the decoder to
make that conversion.

Throughout this module, various parser functions will take a *tokens:
collections.abc.Generator* parameter.  In all cases, *tokens* is
expected to be a *generator iterator* which provides ``pvl.token.Token``
objects.  It should allow for a generated object to be 'returned'
via the generator's send() function.  When parsing the first object
from *tokens*, if an unexpected object is encountered, it will
'return' the object to *tokens*, and raise a ``ValueError``, so
that ``try``-``except`` blocks can be used, and the *generator
iterator* is left in a good state. However, if a parsing anomaly
is discovered deeper in parsing a PVL sequence, then a ``ValueError``
will be thrown into the *tokens* generator iterator (via .throw()).
"""

# Copyright 2015, 2017, 2019-2020, ``pvl`` library authors.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import collections.abc as abc
import re

from .collections import MutableMappingSequence, PVLModule, PVLGroup, PVLObject
from .token import Token
from .grammar import PVLGrammar, OmniGrammar
from .decoder import PVLDecoder, OmniDecoder
from .lexer import lexer as Lexer
from .exceptions import LexerError, ParseError, linecount


class EmptyValueAtLine(str):
    """Empty string to be used as a placeholder for a parameter without
    a value.

    When a label contains a parameter without a value, it is normally
    considered a broken label in PVL. To allow parsing to continue,
    we can rectify the broken parameter-value pair by setting the
    value to have a value of EmptyValueAtLine, which is an empty
    string (and can be treated as such) with some additional properties.

    The argument *lineno* should be the line number of the error from
    the original document, which will be available as a property.

    Examples::
      >>> from pvl.parser import EmptyValueAtLine
      >>> EV1 = EmptyValueAtLine(1)
      >>> EV1
      EmptyValueAtLine(1 does not have a value. Treat as an empty string)
      >>> EV1.lineno
      1
      >>> print(EV1)
      <BLANKLINE>

      >>> EV1 + 'foo'
      'foo'
      >>> # Can be turned into an integer and float as 0:
      >>> int(EV1)
      0
      >>> float(EV1)
      0.0
    """

    def __new__(cls, lineno, *args, **kwargs):
        self = super(EmptyValueAtLine, cls).__new__(cls, "")
        self.lineno = lineno
        return self

    def __int__(self):
        return 0

    def __float__(self):
        return 0.0

    def __repr__(self):
        return (
            "{}({} does not ".format(type(self).__name__, self.lineno)
            + "have a value. Treat as an empty string)"
        )


class PVLParser(object):
    """A parser based on the rules in the CCSDS-641.0-B-2 'Blue Book'
    which defines the PVL language.

    :param grammar: A pvl.grammar object, if None or not specified, it will
                    be set to the grammar parameter of *decoder* (if
                    *decoder* is not None) or will default to
                    :class:`pvl.grammar.OmniGrammar()`.
    :param decoder: defaults to :class:`pvl.decoder.OmniDecoder()`.
    :param lexer_fn: must be a lexer function that takes a ``str``,
        a grammar, and a decoder, as :func:`pvl.lexer.lexer()` does,
        which is the default if none is given.
    :param module_class: must be a subclass of PVLModule, and is the type
        of object that will be returned from this parser's :func:`parse()`
        function.
    :param group_class: must be a subclass of PVLGroup, and is the type
        that will be used to hold PVL elements when a PVL Group is
        encountered during parsing, and must be able to be added to
        via an ``.append()`` function which should take a two-tuple
        of name and value.
    :param object_class: must be a subclass of PVLObject, and is the type
        that will be used to hold PVL elements when a PVL Object is
        encountered during parsing, otherwise similar to *group_class*.
    """

    def __init__(
        self,
        grammar=None,
        decoder=None,
        lexer_fn=None,
        module_class=PVLModule,
        group_class=PVLGroup,
        object_class=PVLObject,
    ):

        self.errors = []
        self.doc = ""

        if lexer_fn is None:
            self.lexer = Lexer
        else:
            self.lexer = lexer_fn

        if grammar is None:
            if decoder is not None:
                self.grammar = decoder.grammar
            else:
                self.grammar = OmniGrammar()
        elif isinstance(grammar, PVLGrammar):
            self.grammar = grammar
        else:
            raise TypeError("The grammar must be an instance of pvl.grammar.")

        if decoder is None:
            self.decoder = OmniDecoder(grammar=self.grammar)
        elif isinstance(decoder, PVLDecoder):
            self.decoder = decoder
        else:
            raise TypeError(
                "The decode must be an instance of pvl.PVLDecoder."
            )

        if issubclass(module_class, MutableMappingSequence):
            self.modcls = module_class
        else:
            raise TypeError(
                "The module_class must be a "
                "pvl.collections.MutableMappingSequence."
            )

        if issubclass(group_class, MutableMappingSequence):
            self.grpcls = group_class
        else:
            raise TypeError(
                "The group_class must be a "
                "pvl.collections.MutableMappingSequence."
            )

        if issubclass(object_class, MutableMappingSequence):
            self.objcls = object_class
        else:
            raise TypeError(
                "The object_class must be a "
                "pvl.collections.MutableMappingSequence."
            )

    def parse(self, s: str):
        """Converts the string, *s* to a PVLModule."""
        self.doc = s
        tokens = self.lexer(s, g=self.grammar, d=self.decoder)
        module = self.parse_module(tokens)
        module.errors = sorted(self.errors)
        return module

    def aggregation_cls(self, begin: str):
        """Returns an initiated object of the group_class or object_class
        as specified on this parser's creation, according to the value
        of *begin*.  If *begin* does not match the Group or Object
        keywords for this parser's grammar, then it will raise a
        ValueError.
        """
        begin_fold = begin.casefold()
        for gk in self.grammar.group_keywords.keys():
            if begin_fold == gk.casefold():
                return self.grpcls()

        for ok in self.grammar.object_keywords.keys():
            if begin_fold == ok.casefold():
                return self.objcls()

        raise ValueError(
            f'The value "{begin}" did not match a Begin '
            "Aggregation Statement."
        )

    def parse_module(self, tokens: abc.Generator):
        """Parses the tokens for a PVL Module.

         <PVL-Module-Contents> ::=
          ( <Assignment-Statement> | <WSC>* | <Aggregation-Block> )*
          [<End-Statement>]

        """
        m = self.modcls()

        parsing = True
        while parsing:
            # print(f'top of while parsing: {m}')
            parsing = False
            for p in (
                self.parse_aggregation_block,
                self.parse_assignment_statement,
                self.parse_end_statement,
            ):
                try:
                    self.parse_WSC_until(None, tokens)
                    # t = next(tokens)
                    # print(f'next token: {t}, {t.pos}')
                    # tokens.send(t)
                    parsed = p(tokens)
                    # print(f'parsed: {parsed}')
                    if parsed is None:  # because parse_end_statement returned
                        return m
                    else:
                        m.append(*parsed)
                        parsing = True
                except LexerError:
                    raise
                except ValueError:
                    pass
            try:
                (m, keep_parsing) = self.parse_module_post_hook(m, tokens)
                if keep_parsing:
                    parsing = True
                else:
                    return m
            except Exception:
                pass

        # print(f'got to bottom: {m}')
        t = next(tokens)
        tokens.throw(
            ValueError,
            "Expecting an Aggregation Block, an Assignment "
            "Statement, or an End Statement, but found "
            f'"{t}" ',
        )

    def parse_module_post_hook(
        self, module: MutableMappingSequence, tokens: abc.Generator
    ):
        """This function is meant to be overridden by subclasses
        that may want to perform some extra processing if
        'normal' parse_module() operations fail to complete.
        See OmniParser for an example.

        This function shall return a two-tuple, with the first item
        being the *module* (altered by processing or unaltered), and
        the second item being a boolean that will signal whether
        the tokens should continue to be parsed to accumulate more
        elements into the returned *module*, or whether the
        *module* is in a good state and should be returned by
        parse_module().

        If the operations within this function are unsuccessful,
        it should raise an exception (any exception descended from
        Exception), which will result in the operation of parse_module()
        as if it were not overridden.
        """
        raise Exception

    def parse_aggregation_block(self, tokens: abc.Generator):
        """Parses the tokens for an Aggregation Block, and returns
        the modcls object that is the result of the parsing and
        decoding.

         <Aggregation-Block> ::= <Begin-Aggegation-Statement>
             (<WSC>* (Assignment-Statement | Aggregation-Block) <WSC>*)+
             <End-Aggregation-Statement>

        The Begin-Aggregation-Statement Name must match the Block-Name
        in the paired End-Aggregation-Statement if a Block-Name is
        present in the End-Aggregation-Statement.
        """
        (begin, block_name) = self.parse_begin_aggregation_statement(tokens)

        agg = self.aggregation_cls(begin)

        while True:
            self.parse_WSC_until(None, tokens)
            try:
                agg.append(*self.parse_aggregation_block(tokens))
            except ValueError:
                try:
                    agg.append(*self.parse_assignment_statement(tokens))
                    # print(f'agg: {agg}')
                    # t = next(tokens)
                    # print(f'next token is: {t}')
                    # tokens.send(t)
                except LexerError:
                    raise
                except ValueError:
                    # t = next(tokens)
                    # print(f'parsing agg block, next token is: {t}')
                    # tokens.send(t)
                    self.parse_end_aggregation(begin, block_name, tokens)
                    break

        return block_name, agg

    def parse_around_equals(self, tokens: abc.Generator) -> None:
        """Parses white space and comments on either side
        of an equals sign.

        *tokens* is expected to be a *generator iterator* which
        provides ``pvl.token`` objects.

        This is shared functionality for Begin Aggregation Statements
        and Assignment Statements.  It basically covers parsing
        anything that has a syntax diagram like this:

          <WSC>* '=' <WSC>*

        """
        if not self.parse_WSC_until("=", tokens):
            try:
                t = next(tokens)
                tokens.send(t)
                raise ValueError(f'Expecting "=", got: {t}')
            except StopIteration:
                raise ParseError('Expecting "=", but ran out of tokens.')

        self.parse_WSC_until(None, tokens)
        return

    def parse_begin_aggregation_statement(
        self, tokens: abc.Generator
    ) -> tuple:
        """Parses the tokens for a Begin Aggregation Statement, and returns
        the name Block Name as a ``str``.

        <Begin-Aggregation-Statement-block> ::=
             <Begin-Aggegation-Statement> <WSC>* '=' <WSC>*
             <Block-Name> [<Statement-Delimiter>]

        Where <Block-Name> ::= <Parameter-Name>

        """
        try:
            begin = next(tokens)
            if not begin.is_begin_aggregation():
                tokens.send(begin)
                raise ValueError(
                    "Expecting a Begin-Aggegation-Statement, but "
                    f"found: {begin}"
                )
        except StopIteration:
            raise ValueError(
                "Ran out of tokens before starting to parse "
                "a Begin-Aggegation-Statement."
            )

        try:
            self.parse_around_equals(tokens)
        except ValueError:
            tokens.throw(
                ValueError, f'Expecting an equals sign after "{begin}" '
            )

        block_name = next(tokens)
        if not block_name.is_parameter_name():
            tokens.throw(
                ValueError,
                f'Expecting a Block-Name after "{begin} =" '
                f'but found: "{block_name}"',
            )

        self.parse_statement_delimiter(tokens)

        return begin, str(block_name)

    def parse_end_aggregation(
        self, begin_agg: str, block_name: str, tokens: abc.Generator
    ) -> None:
        """Parses the tokens for an End Aggregation Statement.

        <End-Aggregation-Statement-block> ::=
             <End-Aggegation-Statement> [<WSC>* '=' <WSC>*
             <Block-Name>] [<Statement-Delimiter>]

        Where <Block-Name> ::= <Parameter-Name>

        """
        end_agg = next(tokens)

        # Need to do a little song and dance to case-independently
        # match the keys:
        for k in self.grammar.aggregation_keywords.keys():
            if k.casefold() == begin_agg.casefold():
                truecase_begin = k
                break
        if (
            end_agg.casefold()
            != self.grammar.aggregation_keywords[truecase_begin].casefold()
        ):
            tokens.send(end_agg)
            raise ValueError(
                "Expecting an End-Aggegation-Statement that "
                "matched the Begin-Aggregation_Statement, "
                f'"{begin_agg}" but found: {end_agg}'
            )

        try:
            self.parse_around_equals(tokens)
        except (ParseError, ValueError):  # No equals statement, which is fine.
            self.parse_statement_delimiter(tokens)
            return None

        t = next(tokens)
        if t != block_name:
            tokens.send(t)
            tokens.throw(
                ValueError,
                f'Expecting a Block-Name after "{end_agg} =" '
                f'that matches "{block_name}", but found: '
                f'"{t}"',
            )

        self.parse_statement_delimiter(tokens)

        return None

    def parse_end_statement(self, tokens: abc.Generator) -> None:
        """Parses the tokens for an End Statement.

        <End-Statement> ::= "END" ( <WSC>* | [<Statement-Delimiter>] )

        """
        try:
            end = next(tokens)
            if not end.is_end_statement():
                tokens.send(end)
                raise ValueError(
                    "Expecting an End Statement, like "
                    f'"{self.grammar.end_statements}" but found '
                    f'"{end}"'
                )

            try:
                t = next(tokens)
                if t.is_WSC():
                    # maybe process comment
                    return
                else:
                    tokens.send(t)
                    return
            except LexerError:
                pass
        except StopIteration:
            pass

        return

    def parse_assignment_statement(self, tokens: abc.Generator) -> tuple:
        """Parses the tokens for an Assignment Statement.

        The returned two-tuple contains the Parameter Name in the
        first element, and the Value in the second.

         <Assignment-Statement> ::= <Parameter-Name> <WSC>* '=' <WSC>*
                                     <Value> [<Statement-Delimiter>]

        """
        try:
            t = next(tokens)
            if t.is_parameter_name():
                parameter_name = str(t)
            else:
                tokens.send(t)
                raise ValueError(
                    "Expecting a Parameter Name, but " f'found: "{t}"'
                )
        except StopIteration:
            raise ValueError(
                "Ran out of tokens before starting to parse "
                "an Assignment-Statement."
            )

        self.parse_around_equals(tokens)

        try:
            # print(f'parameter name: {parameter_name}')
            value = self.parse_value(tokens)
        except StopIteration:
            raise ParseError(
                "Ran out of tokens to parse after the equals "
                "sign in an Assignment-Statement: "
                f'"{parameter_name} =".',
                t,
            )

        self.parse_statement_delimiter(tokens)

        return parameter_name, value

    @staticmethod
    def parse_WSC_until(token: str, tokens: abc.Generator) -> bool:
        """Consumes objects from *tokens*, if the object's *.is_WSC()*
        function returns *True*, it will continue until *token* is
        encountered and will return *True*.  If it encounters an object
        that does not meet these conditions, it will 'return' that
        object to *tokens* and will return *False*.

        *tokens* is expected to be a *generator iterator* which
        provides ``pvl.token`` objects.
        """
        for t in tokens:
            if t == token:
                return True
            elif t.is_WSC():
                # If there's a comment, could parse here.
                pass
            else:
                tokens.send(t)
                return False

    def _parse_set_seq(self, delimiters, tokens: abc.Generator) -> list:
        """The internal parsing of PVL Sets and Sequences are very
        similar, and this function provides that shared logic.

        *delimiters* are a two-tuple containing the start and end
        characters for the PVL Set or Sequence.
        """
        t = next(tokens)
        if t != delimiters[0]:
            tokens.send(t)
            raise ValueError(
                f'Expecting a begin delimiter "{delimiters[0]}" '
                f'but found: "{t}"'
            )
        set_seq = list()
        # Initial WSC and/or empty
        if self.parse_WSC_until(delimiters[1], tokens):
            return set_seq

        # First item:
        set_seq.append(self.parse_value(tokens))
        if self.parse_WSC_until(delimiters[1], tokens):
            return set_seq

        # Remaining items, if any
        for t in tokens:
            # print(f'in loop, t: {t}, set_seq: {set_seq}')
            if t == ",":
                self.parse_WSC_until(None, tokens)  # consume WSC after ','
                set_seq.append(self.parse_value(tokens))
                if self.parse_WSC_until(delimiters[1], tokens):
                    return set_seq
            else:
                tokens.send(t)
                tokens.throw(
                    ValueError,
                    "While parsing, expected a comma (,)" f'but found: "{t}"',
                )

    def parse_set(self, tokens: abc.Generator) -> frozenset:
        """Parses a PVL Set.

         <Set> ::= "{" <WSC>*
                   [ <Value> <WSC>* ( "," <WSC>* <Value> <WSC>* )* ]
                   "}"

        Returns the decoded <Set> as a Python ``frozenset``.  The PVL
        specification doesn't seem to indicate that a PVL Set
        has distinct values (like a Python ``set``), only that the
        ordering of the values is unimportant.  For now, we will
        implement PVL Sets as Python ``frozenset`` objects.

        They are returned as ``frozenset`` objects because PVL Sets
        can contain as their elements other PVL Sets, but since Python
        ``set`` objects are non-hashable, they cannot be members of a set,
        however, ``frozenset`` objects can.
        """
        return frozenset(
            self._parse_set_seq(self.grammar.set_delimiters, tokens)
        )

    def parse_sequence(self, tokens: abc.Generator) -> list:
        """Parses a PVL Sequence.

         <Set> ::= "(" <WSC>*
                   [ <Value> <WSC>* ( "," <WSC>* <Value> <WSC>* )* ]
                   ")"

        Returns the decoded <Sequence> as a Python ``list``.
        """
        return self._parse_set_seq(self.grammar.sequence_delimiters, tokens)

    @staticmethod
    def parse_statement_delimiter(tokens: abc.Generator) -> bool:
        """Parses the tokens for a Statement Delimiter.

        *tokens* is expected to be a *generator iterator* which
        provides ``pvl.token`` objects.

         <Statement-Delimiter> ::= <WSC>*
                     (<white-space-character> | <comment> | ';' | <EOF>)

        Although the above structure comes from Figure 2-4
        of the Blue Book, the <white-space-character> and <comment>
        elements are redundant with the presence of [WSC]*
        so it can be simplified to:

         <Statement-Delimiter> ::= <WSC>* [ ';' | <EOF> ]

        Typically written [<Statement-Delimiter>].
        """
        for t in tokens:
            if t.is_WSC():
                # If there's a comment, could parse here.
                pass
            elif t.is_delimiter():
                return True
            else:
                tokens.send(t)  # Put the next token back into the generator
                return False

    def parse_value(self, tokens: abc.Generator):
        """Parses PVL Values.

         <Value> ::= (<Simple-Value> | <Set> | <Sequence>)
                     [<WSC>* <Units Expression>]

        Returns the decoded <Value> as an appropriate Python object.
        """
        value = None

        try:
            t = next(tokens)
            value = self.decoder.decode_simple_value(t)
        except ValueError:
            tokens.send(t)
            for p in (
                self.parse_set,
                self.parse_sequence,
                self.parse_value_post_hook,
            ):
                try:
                    value = p(tokens)
                    break
                except LexerError:
                    # A LexerError is a subclass of ValueError, but
                    # if we get a LexerError, that's a problem and
                    # we need to raise it, and not let it pass.
                    raise
                except ValueError:
                    # Getting a ValueError is a normal conseqence of
                    # one of the parsing strategies not working,
                    # this pass allows us to go to the next one.
                    pass
            else:
                tokens.throw(
                    ValueError,
                    "Was expecting a Simple Value, or the "
                    "beginning of a Set or Sequence, but "
                    f'found: "{t}"',
                )

        # print(f'in parse_value, value is: {value}')
        self.parse_WSC_until(None, tokens)
        try:
            return self.parse_units(value, tokens)
        except (ValueError, StopIteration):
            return value

    def parse_value_post_hook(self, tokens):
        """This function is meant to be overridden by subclasses
        that may want to perform some extra processing if
        'normal' parse_value() operations fail to yield a value.
        See OmniParser for an example.

        This function shall return an appropriate Python value,
        similar to what parse_value() would return.

        If the operations within this function are unsuccessful,
        it should raise a ValueError which will result in the
        operation of parse_value() as if it were not overridden.
        """
        raise ValueError

    def parse_units(self, value, tokens: abc.Generator) -> str:
        """Parses PVL Units Expression.

         <Units-Expression> ::= "<" [<white-space>] <Units-Value>
                                    [<white-space>] ">"

        and

         <Units-Value> ::= <units-character>
                             [ [ <units-character> | <white-space> ]*
                                 <units-character> ]

        Returns the *value* and the <Units-Value> as a ``Units()``
        object.
        """
        t = next(tokens)

        if not t.startswith(self.grammar.units_delimiters[0]):
            tokens.send(t)
            raise ValueError(
                "Was expecting the start units delimiter, "
                + '"{}" '.format(self.grammar.units_delimiters[0])
                + f'but found "{t}"'
            )

        if not t.endswith(self.grammar.units_delimiters[1]):
            tokens.send(t)
            raise ValueError(
                "Was expecting the end units delimiter, "
                + '"{}" '.format(self.grammar.units_delimiters[1])
                + f'at the end, but found "{t}"'
            )

        delim_strip = t.strip("".join(self.grammar.units_delimiters))

        units_value = delim_strip.strip("".join(self.grammar.whitespace))

        for d in self.grammar.units_delimiters:
            if d in units_value:
                tokens.throw(
                    ValueError,
                    "Was expecting a units character, but found a "
                    f'unit delimiter, "{d}" instead.',
                )

        return self.decoder.decode_quantity(value, units_value)


class ODLParser(PVLParser):
    """A parser based on the rules in the PDS3 Standards Reference
       (version 3.8, 27 Feb 2009) Chapter 12: Object Description
       Language Specification and Usage.

       It extends PVLParser.
    """

    def parse_set(self, tokens: abc.Generator) -> set:
        """Overrides the parent function to return
        the decoded <Set> as a Python ``set``.

        The ODL specification only allows scalar_values in Sets,
        since ODL Sets cannot contain other ODL Sets, an ODL Set
        can be represented as a Python ``set`` (unlike PVL Sets,
        which must be represented as a Python ``frozenset`` objects).
        """
        return set(self._parse_set_seq(self.grammar.set_delimiters, tokens))

    def parse_units(self, value, tokens: abc.Generator) -> str:
        """Extends the parent function, since ODL only allows units
        on numeric values, any others will result in a ValueError.
        """

        if isinstance(value, int) or isinstance(value, float):
            return super().parse_units(value, tokens)

        else:
            raise ValueError(
                "ODL Units Expressions can only follow " "numeric values."
            )


class OmniParser(PVLParser):
    """A permissive PVL/ODL/ISIS label parser that attempts to parse
    all forms of "PVL" that are thrown at it.
    """

    def _empty_value(self, pos):
        eq_pos = self.doc.rfind("=", 0, pos)
        lc = linecount(self.doc, eq_pos)
        self.errors.append(lc)
        return EmptyValueAtLine(lc)

    def parse(self, s: str):
        """Extends the parent function.

        If *any* line ends with a dash (-) followed by a carriage
        return, form-feed, or newline, plus one or more whitespace
        characters on the following line, then those characters, and
        all whitespace characters that begin the next line will
        be removed.
        """
        nodash = re.sub(r"-[\n\r\f]\s*", "", s)
        self.doc = nodash

        return super().parse(nodash)

    def parse_module_post_hook(
        self, module: MutableMappingSequence, tokens: abc.Generator
    ):
        """Overrides the parent function to allow for more
        permissive parsing.  If an Assignment-Statement
        is blank, then the value will be assigned an
        EmptyValueAtLine object.
        """
        # It enables this by checking to see if the next thing is an
        # '=' which means there was an empty assigment at the previous
        # equals sign, and then unwinding the stack to give the
        # previous assignment the EmptyValueAtLine() object and trying
        # to continue parsing.

        # print('in hook')
        try:
            t = next(tokens)
            if t == "=" and len(module) != 0:
                (last_k, last_v) = module[-1]
                last_token = Token(
                    last_v, grammar=self.grammar, decoder=self.decoder
                )
                if last_token.is_parameter_name():
                    # Fix the previous entry
                    module.pop()
                    module.append(last_k, self._empty_value(t.pos))
                    # Now use last_token as the parameter name
                    # for the next assignment, and we must
                    # reproduce the last part of parse-assignment:
                    try:
                        # print(f'parameter name: {last_token}')
                        self.parse_WSC_until(None, tokens)
                        value = self.parse_value(tokens)
                        self.parse_statement_delimiter(tokens)
                        module.append(str(last_token), value)
                    except StopIteration:
                        module.append(
                            str(last_token), self._empty_value(t.pos + 1)
                        )
                        return module, False  # return through parse_module()
                else:
                    tokens.send(t)
            else:
                # The next token isn't an equals sign or the module is
                # empty, so we want return the token and signal
                # parse_module() that it should ignore us.
                tokens.send(t)
                raise Exception

            # Peeking at the next token gives us the opportunity to
            # see if we're at the end of tokens, which we want to handle.
            t = next(tokens)
            tokens.send(t)
            return module, True  # keep parsing
        except StopIteration:
            # If we're out of tokens, that's okay.
            return module, False  # return through parse_module()

    def parse_assignment_statement(self, tokens: abc.Generator) -> tuple:
        """Extends the parent function to allow for more
        permissive parsing.  If an Assignment-Statement
        is blank, then the value will be assigned an
        EmptyValueAtLine object.
        """
        try:
            return super().parse_assignment_statement(tokens)
        except ParseError as err:
            if err.token is not None:
                after_eq = self.doc.find("=", err.token.pos) + 1
                return str(err.token), self._empty_value(after_eq)
            else:
                raise

    def parse_value_post_hook(self, tokens: abc.Generator):
        """Overrides the parent function to allow for more
        permissive parsing.

        If the next token is a reserved word or delimiter,
        then it is returned to the *tokens* and an
        EmptyValueAtLine object is returned as the value.
        """

        t = next(tokens)
        # print(f't: {t}')
        truecase_reserved = [
            x.casefold() for x in self.grammar.reserved_keywords
        ]
        trucase_delim = [x.casefold() for x in self.grammar.delimiters]
        if t.casefold() in (truecase_reserved + trucase_delim):
            # print(f'kw: {kw}')
            # if kw.casefold() == t.casefold():
            # print('match')
            tokens.send(t)
            return self._empty_value(t.pos)
        else:
            raise ValueError
